"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Any, Literal, TYPE_CHECKING, TypeVar, overload
from pandas._libs import lib
from pandas._libs.sparse import BlockIndex, IntIndex, SparseIndex
from pandas._typing import ArrayLike, AstypeArg, Dtype, FillnaOptions, NpDtype, NumpySorter, PositionalIndexer, Scalar, ScalarIndexer, SequenceIndexer, npt
from pandas.core.arraylike import OpsMixin
from pandas.core.arrays import ExtensionArray
from pandas.core.arrays.sparse.dtype import SparseDtype
from pandas.core.base import PandasObject
from enum import Enum
from scipy.sparse import spmatrix
from pandas import Series

"""
SparseArray data structure
"""
if TYPE_CHECKING:
    class ellipsis(Enum):
        Ellipsis = ...
    
    
    Ellipsis = ...
    SparseIndexKind = Literal["integer", "block"]
else:
    ...
SparseArrayT = TypeVar("SparseArrayT", bound="SparseArray")
_sparray_doc_kwargs = ...
class SparseArray(OpsMixin, PandasObject, ExtensionArray):
    """
    An ExtensionArray for storing sparse data.

    Parameters
    ----------
    data : array-like or scalar
        A dense array of values to store in the SparseArray. This may contain
        `fill_value`.
    sparse_index : SparseIndex, optional
    index : Index

        .. deprecated:: 1.4.0
            Use a function like `np.full` to construct an array with the desired
            repeats of the scalar value instead.

    fill_value : scalar, optional
        Elements in data that are ``fill_value`` are not stored in the
        SparseArray. For memory savings, this should be the most common value
        in `data`. By default, `fill_value` depends on the dtype of `data`:

        =========== ==========
        data.dtype  na_value
        =========== ==========
        float       ``np.nan``
        int         ``0``
        bool        False
        datetime64  ``pd.NaT``
        timedelta64 ``pd.NaT``
        =========== ==========

        The fill value is potentially specified in three ways. In order of
        precedence, these are

        1. The `fill_value` argument
        2. ``dtype.fill_value`` if `fill_value` is None and `dtype` is
           a ``SparseDtype``
        3. ``data.dtype.fill_value`` if `fill_value` is None and `dtype`
           is not a ``SparseDtype`` and `data` is a ``SparseArray``.

    kind : str
        Can be 'integer' or 'block', default is 'integer'.
        The type of storage for sparse locations.

        * 'block': Stores a `block` and `block_length` for each
          contiguous *span* of sparse values. This is best when
          sparse data tends to be clumped together, with large
          regions of ``fill-value`` values between sparse values.
        * 'integer': uses an integer to store the location of
          each sparse value.

    dtype : np.dtype or SparseDtype, optional
        The dtype to use for the SparseArray. For numpy dtypes, this
        determines the dtype of ``self.sp_values``. For SparseDtype,
        this determines ``self.sp_values`` and ``self.fill_value``.
    copy : bool, default False
        Whether to explicitly copy the incoming `data` array.

    Attributes
    ----------
    None

    Methods
    -------
    None

    Examples
    --------
    >>> from pandas.arrays import SparseArray
    >>> arr = SparseArray([0, 0, 1, 2])
    >>> arr
    [0, 0, 1, 2]
    Fill: 0
    IntIndex
    Indices: array([2, 3], dtype=int32)
    """
    _subtyp = ...
    _hidden_attrs = ...
    _sparse_index: SparseIndex
    _sparse_values: np.ndarray
    _dtype: SparseDtype
    def __init__(self, data, sparse_index=..., index=..., fill_value=..., kind: SparseIndexKind = ..., dtype: Dtype | None = ..., copy: bool = ...) -> None:
        ...
    
    @classmethod
    def from_spmatrix(cls: type[SparseArrayT], data: spmatrix) -> SparseArrayT:
        """
        Create a SparseArray from a scipy.sparse matrix.

        .. versionadded:: 0.25.0

        Parameters
        ----------
        data : scipy.sparse.sp_matrix
            This should be a SciPy sparse matrix where the size
            of the second dimension is 1. In other words, a
            sparse matrix with a single column.

        Returns
        -------
        SparseArray

        Examples
        --------
        >>> import scipy.sparse
        >>> mat = scipy.sparse.coo_matrix((4, 1))
        >>> pd.arrays.SparseArray.from_spmatrix(mat)
        [0.0, 0.0, 0.0, 0.0]
        Fill: 0.0
        IntIndex
        Indices: array([], dtype=int32)
        """
        ...
    
    def __array__(self, dtype: NpDtype | None = ...) -> np.ndarray:
        ...
    
    def __setitem__(self, key, value):
        ...
    
    @property
    def sp_index(self) -> SparseIndex:
        """
        The SparseIndex containing the location of non- ``fill_value`` points.
        """
        ...
    
    @property
    def sp_values(self) -> np.ndarray:
        """
        An ndarray containing the non- ``fill_value`` values.

        Examples
        --------
        >>> s = SparseArray([0, 0, 1, 0, 2], fill_value=0)
        >>> s.sp_values
        array([1, 2])
        """
        ...
    
    @property
    def dtype(self) -> SparseDtype:
        ...
    
    @property
    def fill_value(self): # -> Any:
        """
        Elements in `data` that are `fill_value` are not stored.

        For memory savings, this should be the most common value in the array.
        """
        ...
    
    @fill_value.setter
    def fill_value(self, value): # -> None:
        ...
    
    @property
    def kind(self) -> SparseIndexKind:
        """
        The kind of sparse index for this array. One of {'integer', 'block'}.
        """
        ...
    
    def __len__(self) -> int:
        ...
    
    @property
    def nbytes(self) -> int:
        ...
    
    @property
    def density(self) -> float:
        """
        The percent of non- ``fill_value`` points, as decimal.

        Examples
        --------
        >>> s = SparseArray([0, 0, 1, 1, 1], fill_value=0)
        >>> s.density
        0.6
        """
        ...
    
    @property
    def npoints(self) -> int:
        """
        The number of non- ``fill_value`` points.

        Examples
        --------
        >>> s = SparseArray([0, 0, 1, 1, 1], fill_value=0)
        >>> s.npoints
        3
        """
        ...
    
    def isna(self): # -> Self@SparseArray:
        ...
    
    def fillna(self: SparseArrayT, value=..., method: FillnaOptions | None = ..., limit: int | None = ...) -> SparseArrayT:
        """
        Fill missing values with `value`.

        Parameters
        ----------
        value : scalar, optional
        method : str, optional

            .. warning::

               Using 'method' will result in high memory use,
               as all `fill_value` methods will be converted to
               an in-memory ndarray

        limit : int, optional

        Returns
        -------
        SparseArray

        Notes
        -----
        When `value` is specified, the result's ``fill_value`` depends on
        ``self.fill_value``. The goal is to maintain low-memory use.

        If ``self.fill_value`` is NA, the result dtype will be
        ``SparseDtype(self.dtype, fill_value=value)``. This will preserve
        amount of memory used before and after filling.

        When ``self.fill_value`` is not NA, the result dtype will be
        ``self.dtype``. Again, this preserves the amount of memory used.
        """
        ...
    
    def shift(self: SparseArrayT, periods: int = ..., fill_value=...) -> SparseArrayT:
        ...
    
    def unique(self: SparseArrayT) -> SparseArrayT:
        ...
    
    def factorize(self, na_sentinel: int | lib.NoDefault = ..., use_na_sentinel: bool | lib.NoDefault = ...) -> tuple[np.ndarray, SparseArray]:
        ...
    
    def value_counts(self, dropna: bool = ...) -> Series:
        """
        Returns a Series containing counts of unique values.

        Parameters
        ----------
        dropna : bool, default True
            Don't include counts of NaN, even if NaN is in sp_values.

        Returns
        -------
        counts : Series
        """
        ...
    
    @overload
    def __getitem__(self, key: ScalarIndexer) -> Any:
        ...
    
    @overload
    def __getitem__(self: SparseArrayT, key: SequenceIndexer | tuple[int | ellipsis, ...]) -> SparseArrayT:
        ...
    
    def __getitem__(self: SparseArrayT, key: PositionalIndexer | tuple[int | ellipsis, ...]) -> SparseArrayT | Any:
        ...
    
    def take(self: SparseArrayT, indices, *, allow_fill: bool = ..., fill_value=...) -> SparseArrayT:
        ...
    
    def searchsorted(self, v: ArrayLike | object, side: Literal["left", "right"] = ..., sorter: NumpySorter = ...) -> npt.NDArray[np.intp] | np.intp:
        ...
    
    def copy(self: SparseArrayT) -> SparseArrayT:
        ...
    
    def astype(self, dtype: AstypeArg | None = ..., copy: bool = ...): # -> Self@SparseArray:
        """
        Change the dtype of a SparseArray.

        The output will always be a SparseArray. To convert to a dense
        ndarray with a certain dtype, use :meth:`numpy.asarray`.

        Parameters
        ----------
        dtype : np.dtype or ExtensionDtype
            For SparseDtype, this changes the dtype of
            ``self.sp_values`` and the ``self.fill_value``.

            For other dtypes, this only changes the dtype of
            ``self.sp_values``.

        copy : bool, default True
            Whether to ensure a copy is made, even if not necessary.

        Returns
        -------
        SparseArray

        Examples
        --------
        >>> arr = pd.arrays.SparseArray([0, 0, 1, 2])
        >>> arr
        [0, 0, 1, 2]
        Fill: 0
        IntIndex
        Indices: array([2, 3], dtype=int32)

        >>> arr.astype(SparseDtype(np.dtype('int32')))
        [0, 0, 1, 2]
        Fill: 0
        IntIndex
        Indices: array([2, 3], dtype=int32)

        Using a NumPy dtype with a different kind (e.g. float) will coerce
        just ``self.sp_values``.

        >>> arr.astype(SparseDtype(np.dtype('float64')))
        ... # doctest: +NORMALIZE_WHITESPACE
        [nan, nan, 1.0, 2.0]
        Fill: nan
        IntIndex
        Indices: array([2, 3], dtype=int32)

        Using a SparseDtype, you can also change the fill value as well.

        >>> arr.astype(SparseDtype("float64", fill_value=0.0))
        ... # doctest: +NORMALIZE_WHITESPACE
        [0.0, 0.0, 1.0, 2.0]
        Fill: 0.0
        IntIndex
        Indices: array([2, 3], dtype=int32)
        """
        ...
    
    def map(self: SparseArrayT, mapper) -> SparseArrayT:
        """
        Map categories using an input mapping or function.

        Parameters
        ----------
        mapper : dict, Series, callable
            The correspondence from old values to new.

        Returns
        -------
        SparseArray
            The output array will have the same density as the input.
            The output fill value will be the result of applying the
            mapping to ``self.fill_value``

        Examples
        --------
        >>> arr = pd.arrays.SparseArray([0, 1, 2])
        >>> arr.map(lambda x: x + 10)
        [10, 11, 12]
        Fill: 10
        IntIndex
        Indices: array([1, 2], dtype=int32)

        >>> arr.map({0: 10, 1: 11, 2: 12})
        [10, 11, 12]
        Fill: 10
        IntIndex
        Indices: array([1, 2], dtype=int32)

        >>> arr.map(pd.Series([10, 11, 12], index=[0, 1, 2]))
        [10, 11, 12]
        Fill: 10
        IntIndex
        Indices: array([1, 2], dtype=int32)
        """
        ...
    
    def to_dense(self) -> np.ndarray:
        """
        Convert SparseArray to a NumPy array.

        Returns
        -------
        arr : NumPy array
        """
        ...
    
    def __setstate__(self, state) -> None:
        """Necessary for making this object picklable"""
        ...
    
    def nonzero(self) -> tuple[npt.NDArray[np.int32]]:
        ...
    
    def all(self, axis=..., *args, **kwargs): # -> bool_ | Literal[False]:
        """
        Tests whether all elements evaluate True

        Returns
        -------
        all : bool

        See Also
        --------
        numpy.all
        """
        ...
    
    def any(self, axis=..., *args, **kwargs): # -> bool:
        """
        Tests whether at least one of elements evaluate True

        Returns
        -------
        any : bool

        See Also
        --------
        numpy.any
        """
        ...
    
    def sum(self, axis: int = ..., min_count: int = ..., skipna: bool = ..., *args, **kwargs) -> Scalar:
        """
        Sum of non-NA/null values

        Parameters
        ----------
        axis : int, default 0
            Not Used. NumPy compatibility.
        min_count : int, default 0
            The required number of valid values to perform the summation. If fewer
            than ``min_count`` valid values are present, the result will be the missing
            value indicator for subarray type.
        *args, **kwargs
            Not Used. NumPy compatibility.

        Returns
        -------
        scalar
        """
        ...
    
    def cumsum(self, axis: int = ..., *args, **kwargs) -> SparseArray:
        """
        Cumulative sum of non-NA/null values.

        When performing the cumulative summation, any non-NA/null values will
        be skipped. The resulting SparseArray will preserve the locations of
        NaN values, but the fill value will be `np.nan` regardless.

        Parameters
        ----------
        axis : int or None
            Axis over which to perform the cumulative summation. If None,
            perform cumulative summation over flattened array.

        Returns
        -------
        cumsum : SparseArray
        """
        ...
    
    def mean(self, axis=..., *args, **kwargs): # -> Any:
        """
        Mean of non-NA/null values

        Returns
        -------
        mean : float
        """
        ...
    
    def max(self, *, axis: int | None = ..., skipna: bool = ...): # -> Scalar:
        """
        Max of array values, ignoring NA values if specified.

        Parameters
        ----------
        axis : int, default 0
            Not Used. NumPy compatibility.
        skipna : bool, default True
            Whether to ignore NA values.

        Returns
        -------
        scalar
        """
        ...
    
    def min(self, *, axis: int | None = ..., skipna: bool = ...): # -> Scalar:
        """
        Min of array values, ignoring NA values if specified.

        Parameters
        ----------
        axis : int, default 0
            Not Used. NumPy compatibility.
        skipna : bool, default True
            Whether to ignore NA values.

        Returns
        -------
        scalar
        """
        ...
    
    def argmax(self, skipna: bool = ...) -> int:
        ...
    
    def argmin(self, skipna: bool = ...) -> int:
        ...
    
    _HANDLED_TYPES = ...
    def __array_ufunc__(self, ufunc: np.ufunc, method: str, *inputs, **kwargs): # -> _NotImplementedType | tuple[Unknown, ...] | Any | tuple[Self@SparseArray, ...] | Self@SparseArray | None:
        ...
    
    _logical_method = ...
    def __pos__(self) -> SparseArray:
        ...
    
    def __neg__(self) -> SparseArray:
        ...
    
    def __invert__(self) -> SparseArray:
        ...
    
    def __abs__(self) -> SparseArray:
        ...
    
    def __repr__(self) -> str:
        ...
    


def make_sparse(arr: np.ndarray, kind: SparseIndexKind = ..., fill_value=..., dtype: NpDtype | None = ...): # -> tuple[Unknown | ndarray[Any, Unknown] | ndarray[Any, dtype[Any]], IntIndex | BlockIndex, object | Unknown | float | Literal[0, False]]:
    """
    Convert ndarray to sparse format

    Parameters
    ----------
    arr : ndarray
    kind : {'block', 'integer'}
    fill_value : NaN or another value
    dtype : np.dtype, optional
    copy : bool, default False

    Returns
    -------
    (sparse_values, index, fill_value) : (ndarray, SparseIndex, Scalar)
    """
    ...

@overload
def make_sparse_index(length: int, indices, kind: Literal["block"]) -> BlockIndex:
    ...

@overload
def make_sparse_index(length: int, indices, kind: Literal["integer"]) -> IntIndex:
    ...

def make_sparse_index(length: int, indices, kind: SparseIndexKind) -> SparseIndex:
    ...




"""
This type stub file was generated by pyright.
"""

from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    ...
class ParserBase:
    class BadLineHandleMethod(Enum):
        ERROR = ...
        WARN = ...
        SKIP = ...
    
    
    _implicit_index: bool = ...
    _first_chunk: bool
    def __init__(self, kwds) -> None:
        ...
    
    def close(self) -> None:
        ...
    


parser_defaults = ...
def is_index_col(col) -> bool:
    ...




"""
This type stub file was generated by pyright.
"""

from typing import TYPE_CHECKING

""" define generic base classes for pandas objects """
if TYPE_CHECKING:
    ...
def create_pandas_abc_type(name, attr, comp): # -> Type[__class_ABCBase]:
    ...

ABCInt64Index = ...
ABCUInt64Index = ...
ABCRangeIndex = ...
ABCFloat64Index = ...
ABCMultiIndex = ...
ABCDatetimeIndex = ...
ABCTimedeltaIndex = ...
ABCPeriodIndex = ...
ABCCategoricalIndex = ...
ABCIntervalIndex = ...
ABCIndex = ...
ABCNDFrame = ...
ABCSeries = ...
ABCDataFrame = ...
ABCCategorical = ...
ABCDatetimeArray = ...
ABCTimedeltaArray = ...
ABCPeriodArray = ...
ABCExtensionArray = ...
ABCPandasArray = ...



"""
This type stub file was generated by pyright.
"""

from typing import Any, TYPE_CHECKING
from pandas._typing import CompressionOptions, FilePath, ReadBuffer, StorageOptions, WriteBuffer
from pandas.util._decorators import doc
from pandas.core.shared_docs import _shared_docs
from pandas import DataFrame

"""
:mod:`pandas.io.formats.xml` is a module for formatting data in XML.
"""
if TYPE_CHECKING:
    ...
@doc(storage_options=_shared_docs["storage_options"], compression_options=_shared_docs["compression_options"] % "path_or_buffer")
class BaseXMLFormatter:
    """
    Subclass for formatting data in XML.

    Parameters
    ----------
    path_or_buffer : str or file-like
        This can be either a string of raw XML, a valid URL,
        file or file-like object.

    index : bool
        Whether to include index in xml document.

    row_name : str
        Name for root of xml document. Default is 'data'.

    root_name : str
        Name for row elements of xml document. Default is 'row'.

    na_rep : str
        Missing data representation.

    attrs_cols : list
        List of columns to write as attributes in row element.

    elem_cols : list
        List of columns to write as children in row element.

    namespaces : dict
        The namespaces to define in XML document as dicts with key
        being namespace and value the URI.

    prefix : str
        The prefix for each element in XML document including root.

    encoding : str
        Encoding of xml object or document.

    xml_declaration : bool
        Whether to include xml declaration at top line item in xml.

    pretty_print : bool
        Whether to write xml document with line breaks and indentation.

    stylesheet : str or file-like
        A URL, file, file-like object, or a raw string containing XSLT.

    {compression_options}

        .. versionchanged:: 1.4.0 Zstandard support.

    {storage_options}

    See also
    --------
    pandas.io.formats.xml.EtreeXMLFormatter
    pandas.io.formats.xml.LxmlXMLFormatter

    """
    def __init__(self, frame: DataFrame, path_or_buffer: FilePath | WriteBuffer[bytes] | WriteBuffer[str] | None = ..., index: bool = ..., root_name: str | None = ..., row_name: str | None = ..., na_rep: str | None = ..., attr_cols: list[str] | None = ..., elem_cols: list[str] | None = ..., namespaces: dict[str | None, str] | None = ..., prefix: str | None = ..., encoding: str = ..., xml_declaration: bool | None = ..., pretty_print: bool | None = ..., stylesheet: FilePath | ReadBuffer[str] | ReadBuffer[bytes] | None = ..., compression: CompressionOptions = ..., storage_options: StorageOptions = ...) -> None:
        ...
    
    def build_tree(self) -> bytes:
        """
        Build tree from  data.

        This method initializes the root and builds attributes and elements
        with optional namespaces.
        """
        ...
    
    def validate_columns(self) -> None:
        """
        Validate elems_cols and attrs_cols.

        This method will check if columns is list-like.

        Raises
        ------
        ValueError
            * If value is not a list and less then length of nodes.
        """
        ...
    
    def validate_encoding(self) -> None:
        """
        Validate encoding.

        This method will check if encoding is among listed under codecs.

        Raises
        ------
        LookupError
            * If encoding is not available in codecs.
        """
        ...
    
    def process_dataframe(self) -> dict[int | str, dict[str, Any]]:
        """
        Adjust Data Frame to fit xml output.

        This method will adjust underlying data frame for xml output,
        including optionally replacing missing values and including indexes.
        """
        ...
    
    def handle_indexes(self) -> None:
        """
        Handle indexes.

        This method will add indexes into attr_cols or elem_cols.
        """
        ...
    
    def get_prefix_uri(self) -> str:
        """
        Get uri of namespace prefix.

        This method retrieves corresponding URI to prefix in namespaces.

        Raises
        ------
        KeyError
            *If prefix is not included in namespace dict.
        """
        ...
    
    def other_namespaces(self) -> dict:
        """
        Define other namespaces.

        This method will build dictionary of namespaces attributes
        for root element, conditionally with optional namespaces and
        prefix.
        """
        ...
    
    def build_attribs(self, d: dict[str, Any], elem_row: Any) -> Any:
        """
        Create attributes of row.

        This method adds attributes using attr_cols to row element and
        works with tuples for multindex or hierarchical columns.
        """
        ...
    
    def build_elems(self, d: dict[str, Any], elem_row: Any) -> None:
        """
        Create child elements of row.

        This method adds child elements using elem_cols to row element and
        works with tuples for multindex or hierarchical columns.
        """
        ...
    
    def write_output(self) -> str | None:
        ...
    


class EtreeXMLFormatter(BaseXMLFormatter):
    """
    Class for formatting data in xml using Python standard library
    modules: `xml.etree.ElementTree` and `xml.dom.minidom`.
    """
    def build_tree(self) -> bytes:
        ...
    
    def get_prefix_uri(self) -> str:
        ...
    
    def build_elems(self, d: dict[str, Any], elem_row: Any) -> None:
        ...
    
    def prettify_tree(self) -> bytes:
        """
        Output tree for pretty print format.

        This method will pretty print xml with line breaks and indentation.
        """
        ...
    
    def add_declaration(self) -> bytes:
        """
        Add xml declaration.

        This method will add xml declaration of working tree. Currently,
        xml_declaration is supported in etree starting in Python 3.8.
        """
        ...
    
    def remove_declaration(self) -> bytes:
        """
        Remove xml declaration.

        This method will remove xml declaration of working tree. Currently,
        pretty_print is not supported in etree.
        """
        ...
    


class LxmlXMLFormatter(BaseXMLFormatter):
    """
    Class for formatting data in xml using Python standard library
    modules: `xml.etree.ElementTree` and `xml.dom.minidom`.
    """
    def __init__(self, *args, **kwargs) -> None:
        ...
    
    def build_tree(self) -> bytes:
        """
        Build tree from  data.

        This method initializes the root and builds attributes and elements
        with optional namespaces.
        """
        ...
    
    def convert_empty_str_key(self) -> None:
        """
        Replace zero-length string in `namespaces`.

        This method will replace '' with None to align to `lxml`
        requirement that empty string prefixes are not allowed.
        """
        ...
    
    def get_prefix_uri(self) -> str:
        ...
    
    def build_elems(self, d: dict[str, Any], elem_row: Any) -> None:
        ...
    
    def transform_doc(self) -> bytes:
        """
        Parse stylesheet from file or buffer and run it.

        This method will parse stylesheet object into tree for parsing
        conditionally by its specific object type, then transforms
        original tree with XSLT script.
        """
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import TYPE_CHECKING
from pandas.core.generic import NDFrame

"""
Module containing utilities for NDFrame.sample() and .GroupBy.sample()
"""
if TYPE_CHECKING:
    ...
def preprocess_weights(obj: NDFrame, weights, axis: int) -> np.ndarray:
    """
    Process and validate the `weights` argument to `NDFrame.sample` and
    `.GroupBy.sample`.

    Returns `weights` as an ndarray[np.float64], validated except for normalizing
    weights (because that must be done groupwise in groupby sampling).
    """
    ...

def process_sampling_size(n: int | None, frac: float | None, replace: bool) -> int | None:
    """
    Process and validate the `n` and `frac` arguments to `NDFrame.sample` and
    `.GroupBy.sample`.

    Returns None if `frac` should be used (variable sampling sizes), otherwise returns
    the constant sampling size.
    """
    ...

def sample(obj_len: int, size: int, replace: bool, weights: np.ndarray | None, random_state: np.random.RandomState | np.random.Generator) -> np.ndarray:
    """
    Randomly sample `size` indices in `np.arange(obj_len)`

    Parameters
    ----------
    obj_len : int
        The length of the indices being considered
    size : int
        The number of values to choose
    replace : bool
        Allow or disallow sampling of the same row more than once.
    weights : np.ndarray[np.float64] or None
        If None, equal probability weighting, otherwise weights according
        to the vector normalized
    random_state: np.random.RandomState or np.random.Generator
        State used for the random sampling

    Returns
    -------
    np.ndarray[np.intp]
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Any, Callable, Hashable, Iterator, TYPE_CHECKING
from pandas._libs import lib
from pandas._typing import Dtype, npt
from pandas.util._decorators import cache_readonly, doc
from pandas.core.indexes.numeric import Int64Index, NumericIndex
from pandas.core.ops.common import unpack_zerodim_and_defer
from pandas import Index

if TYPE_CHECKING:
    ...
_empty_range = ...
class RangeIndex(NumericIndex):
    """
    Immutable Index implementing a monotonic integer range.

    RangeIndex is a memory-saving special case of Int64Index limited to
    representing monotonic ranges. Using RangeIndex may in some instances
    improve computing speed.

    This is the default index type used
    by DataFrame and Series when no explicit index is provided by the user.

    Parameters
    ----------
    start : int (default: 0), range, or other RangeIndex instance
        If int and "stop" is not given, interpreted as "stop" instead.
    stop : int (default: 0)
    step : int (default: 1)
    dtype : np.int64
        Unused, accepted for homogeneity with other index types.
    copy : bool, default False
        Unused, accepted for homogeneity with other index types.
    name : object, optional
        Name to be stored in the index.

    Attributes
    ----------
    start
    stop
    step

    Methods
    -------
    from_range

    See Also
    --------
    Index : The base pandas Index type.
    Int64Index : Index of int64 data.
    """
    _typ = ...
    _dtype_validation_metadata = ...
    _range: range
    _is_backward_compat_public_numeric_index: bool = ...
    def __new__(cls, start=..., stop=..., step=..., dtype: Dtype | None = ..., copy: bool = ..., name: Hashable = ...) -> RangeIndex:
        ...
    
    @classmethod
    def from_range(cls, data: range, name=..., dtype: Dtype | None = ...) -> RangeIndex:
        """
        Create RangeIndex from a range object.

        Returns
        -------
        RangeIndex
        """
        ...
    
    def __reduce__(self): # -> tuple[(cls: Unknown, d: Unknown) -> (ABCPeriodIndex | MultiIndex | Unknown), tuple[Type[Self@RangeIndex], dict[str, Hashable]], None]:
        ...
    
    _deprecation_message = ...
    @property
    def start(self) -> int:
        """
        The value of the `start` parameter (``0`` if this was not supplied).
        """
        ...
    
    @property
    def stop(self) -> int:
        """
        The value of the `stop` parameter.
        """
        ...
    
    @property
    def step(self) -> int:
        """
        The value of the `step` parameter (``1`` if this was not supplied).
        """
        ...
    
    @cache_readonly
    def nbytes(self) -> int:
        """
        Return the number of bytes in the underlying data.
        """
        ...
    
    def memory_usage(self, deep: bool = ...) -> int:
        """
        Memory usage of my values

        Parameters
        ----------
        deep : bool
            Introspect the data deeply, interrogate
            `object` dtypes for system-level memory consumption

        Returns
        -------
        bytes used

        Notes
        -----
        Memory usage does not include memory consumed by elements that
        are not components of the array if deep=False

        See Also
        --------
        numpy.ndarray.nbytes
        """
        ...
    
    @property
    def dtype(self) -> np.dtype:
        ...
    
    @property
    def is_unique(self) -> bool:
        """return if the index has unique values"""
        ...
    
    @cache_readonly
    def is_monotonic_increasing(self) -> bool:
        ...
    
    @cache_readonly
    def is_monotonic_decreasing(self) -> bool:
        ...
    
    def __contains__(self, key: Any) -> bool:
        ...
    
    @property
    def inferred_type(self) -> str:
        ...
    
    @doc(Int64Index.get_loc)
    def get_loc(self, key, method=..., tolerance=...): # -> int | slice | ndarray[Unknown, Unknown]:
        ...
    
    def tolist(self) -> list[int]:
        ...
    
    @doc(Int64Index.__iter__)
    def __iter__(self) -> Iterator[int]:
        ...
    
    @doc(Int64Index.copy)
    def copy(self, name: Hashable = ..., deep: bool = ..., dtype: Dtype | None = ..., names=...): # -> Self@RangeIndex | Index:
        ...
    
    def min(self, axis=..., skipna: bool = ..., *args, **kwargs) -> int:
        """The minimum value of the RangeIndex"""
        ...
    
    def max(self, axis=..., skipna: bool = ..., *args, **kwargs) -> int:
        """The maximum value of the RangeIndex"""
        ...
    
    def argsort(self, *args, **kwargs) -> npt.NDArray[np.intp]:
        """
        Returns the indices that would sort the index and its
        underlying data.

        Returns
        -------
        np.ndarray[np.intp]

        See Also
        --------
        numpy.ndarray.argsort
        """
        ...
    
    def factorize(self, sort: bool = ..., na_sentinel: int | lib.NoDefault = ..., use_na_sentinel: bool | lib.NoDefault = ...) -> tuple[npt.NDArray[np.intp], RangeIndex]:
        ...
    
    def equals(self, other: object) -> bool:
        """
        Determines if two Index objects contain the same elements.
        """
        ...
    
    def sort_values(self, return_indexer: bool = ..., ascending: bool = ..., na_position: str = ..., key: Callable | None = ...): # -> tuple[RangeIndex | MultiIndex, NDArray[intp] | Unknown | Any] | RangeIndex | MultiIndex | tuple[RangeIndex | int | Any | Unknown | ExtensionArray | Self@RangeIndex, RangeIndex | int | Any | Unknown | ExtensionArray] | int | Any | ExtensionArray | Self@RangeIndex:
        ...
    
    def symmetric_difference(self, other, result_name: Hashable = ..., sort=...): # -> ABCMultiIndex | RangeIndex | Index | Any | Self@Index | None:
        ...
    
    def delete(self, loc) -> Index:
        ...
    
    def insert(self, loc: int, item) -> Index:
        ...
    
    def __len__(self) -> int:
        """
        return the length of the RangeIndex
        """
        ...
    
    @property
    def size(self) -> int:
        ...
    
    def __getitem__(self, key): # -> RangeIndex | int | Any | ExtensionArray:
        """
        Conserve RangeIndex type for scalar and slice keys.
        """
        ...
    
    @unpack_zerodim_and_defer("__floordiv__")
    def __floordiv__(self, other): # -> RangeIndex | _NotImplementedType:
        ...
    
    def all(self, *args, **kwargs) -> bool:
        ...
    
    def any(self, *args, **kwargs) -> bool:
        ...
    





"""
This type stub file was generated by pyright.
"""

import textwrap
import numpy as np
import pandas._libs.lib as lib
from typing import Any, Generic, Hashable, Literal, TYPE_CHECKING, TypeVar, final, overload
from pandas._typing import ArrayLike, DtypeObj, IndexLabel, NDFrameT, NumpySorter, NumpyValueArrayLike, Shape, npt
from pandas.util._decorators import cache_readonly, doc
from pandas.core import algorithms
from pandas.core.accessor import DirNamesMixin
from pandas.core.arraylike import OpsMixin
from pandas.core.arrays import ExtensionArray
from pandas import Series

"""
Base and utility classes for pandas objects.
"""
if TYPE_CHECKING:
    ...
_shared_docs: dict[str, str] = ...
_indexops_doc_kwargs = ...
_T = TypeVar("_T", bound="IndexOpsMixin")
class PandasObject(DirNamesMixin):
    """
    Baseclass for various pandas objects.
    """
    _cache: dict[str, Any]
    def __repr__(self) -> str:
        """
        Return a string representation for a particular object.
        """
        ...
    
    def __sizeof__(self) -> int:
        """
        Generates the total memory usage for an object that returns
        either a value or Series of values
        """
        ...
    


class NoNewAttributesMixin:
    """
    Mixin which prevents adding new attributes.

    Prevents additional attributes via xxx.attribute = "something" after a
    call to `self.__freeze()`. Mainly used to prevent the user from using
    wrong attributes on an accessor (`Series.cat/.str/.dt`).

    If you really want to add a new attribute at a later time, you need to use
    `object.__setattr__(self, key, value)`.
    """
    def __setattr__(self, key: str, value) -> None:
        ...
    


class SelectionMixin(Generic[NDFrameT]):
    """
    mixin implementing the selection & aggregation interface on a group-like
    object sub-classes need to define: obj, exclusions
    """
    obj: NDFrameT
    _selection: IndexLabel | None = ...
    exclusions: frozenset[Hashable]
    _internal_names = ...
    _internal_names_set = ...
    @final
    @cache_readonly
    def ndim(self) -> int:
        ...
    
    def __getitem__(self, key):
        ...
    
    def aggregate(self, func, *args, **kwargs):
        ...
    
    agg = ...


class IndexOpsMixin(OpsMixin):
    """
    Common ops mixin to support a unified interface / docs for Series / Index
    """
    __array_priority__ = ...
    _hidden_attrs: frozenset[str] = ...
    @property
    def dtype(self) -> DtypeObj:
        ...
    
    def transpose(self: _T, *args, **kwargs) -> _T:
        """
        Return the transpose, which is by definition self.

        Returns
        -------
        %(klass)s
        """
        ...
    
    T = ...
    @property
    def shape(self) -> Shape:
        """
        Return a tuple of the shape of the underlying data.
        """
        ...
    
    def __len__(self) -> int:
        ...
    
    @property
    def ndim(self) -> Literal[1]:
        """
        Number of dimensions of the underlying data, by definition 1.
        """
        ...
    
    def item(self): # -> Any:
        """
        Return the first element of the underlying data as a Python scalar.

        Returns
        -------
        scalar
            The first element of %(klass)s.

        Raises
        ------
        ValueError
            If the data is not length-1.
        """
        ...
    
    @property
    def nbytes(self) -> int:
        """
        Return the number of bytes in the underlying data.
        """
        ...
    
    @property
    def size(self) -> int:
        """
        Return the number of elements in the underlying data.
        """
        ...
    
    @property
    def array(self) -> ExtensionArray:
        """
        The ExtensionArray of the data backing this Series or Index.

        Returns
        -------
        ExtensionArray
            An ExtensionArray of the values stored within. For extension
            types, this is the actual array. For NumPy native types, this
            is a thin (no copy) wrapper around :class:`numpy.ndarray`.

            ``.array`` differs ``.values`` which may require converting the
            data to a different form.

        See Also
        --------
        Index.to_numpy : Similar method that always returns a NumPy array.
        Series.to_numpy : Similar method that always returns a NumPy array.

        Notes
        -----
        This table lays out the different array types for each extension
        dtype within pandas.

        ================== =============================
        dtype              array type
        ================== =============================
        category           Categorical
        period             PeriodArray
        interval           IntervalArray
        IntegerNA          IntegerArray
        string             StringArray
        boolean            BooleanArray
        datetime64[ns, tz] DatetimeArray
        ================== =============================

        For any 3rd-party extension types, the array type will be an
        ExtensionArray.

        For all remaining dtypes ``.array`` will be a
        :class:`arrays.NumpyExtensionArray` wrapping the actual ndarray
        stored within. If you absolutely need a NumPy array (possibly with
        copying / coercing data), then use :meth:`Series.to_numpy` instead.

        Examples
        --------
        For regular NumPy types like int, and float, a PandasArray
        is returned.

        >>> pd.Series([1, 2, 3]).array
        <PandasArray>
        [1, 2, 3]
        Length: 3, dtype: int64

        For extension types, like Categorical, the actual ExtensionArray
        is returned

        >>> ser = pd.Series(pd.Categorical(['a', 'b', 'a']))
        >>> ser.array
        ['a', 'b', 'a']
        Categories (2, object): ['a', 'b']
        """
        ...
    
    def to_numpy(self, dtype: npt.DTypeLike | None = ..., copy: bool = ..., na_value: object = ..., **kwargs) -> np.ndarray:
        """
        A NumPy ndarray representing the values in this Series or Index.

        Parameters
        ----------
        dtype : str or numpy.dtype, optional
            The dtype to pass to :meth:`numpy.asarray`.
        copy : bool, default False
            Whether to ensure that the returned value is not a view on
            another array. Note that ``copy=False`` does not *ensure* that
            ``to_numpy()`` is no-copy. Rather, ``copy=True`` ensure that
            a copy is made, even if not strictly necessary.
        na_value : Any, optional
            The value to use for missing values. The default value depends
            on `dtype` and the type of the array.

            .. versionadded:: 1.0.0

        **kwargs
            Additional keywords passed through to the ``to_numpy`` method
            of the underlying array (for extension arrays).

            .. versionadded:: 1.0.0

        Returns
        -------
        numpy.ndarray

        See Also
        --------
        Series.array : Get the actual data stored within.
        Index.array : Get the actual data stored within.
        DataFrame.to_numpy : Similar method for DataFrame.

        Notes
        -----
        The returned array will be the same up to equality (values equal
        in `self` will be equal in the returned array; likewise for values
        that are not equal). When `self` contains an ExtensionArray, the
        dtype may be different. For example, for a category-dtype Series,
        ``to_numpy()`` will return a NumPy array and the categorical dtype
        will be lost.

        For NumPy dtypes, this will be a reference to the actual data stored
        in this Series or Index (assuming ``copy=False``). Modifying the result
        in place will modify the data stored in the Series or Index (not that
        we recommend doing that).

        For extension types, ``to_numpy()`` *may* require copying data and
        coercing the result to a NumPy type (possibly object), which may be
        expensive. When you need a no-copy reference to the underlying data,
        :attr:`Series.array` should be used instead.

        This table lays out the different dtypes and default return types of
        ``to_numpy()`` for various dtypes within pandas.

        ================== ================================
        dtype              array type
        ================== ================================
        category[T]        ndarray[T] (same dtype as input)
        period             ndarray[object] (Periods)
        interval           ndarray[object] (Intervals)
        IntegerNA          ndarray[object]
        datetime64[ns]     datetime64[ns]
        datetime64[ns, tz] ndarray[object] (Timestamps)
        ================== ================================

        Examples
        --------
        >>> ser = pd.Series(pd.Categorical(['a', 'b', 'a']))
        >>> ser.to_numpy()
        array(['a', 'b', 'a'], dtype=object)

        Specify the `dtype` to control how datetime-aware data is represented.
        Use ``dtype=object`` to return an ndarray of pandas :class:`Timestamp`
        objects, each with the correct ``tz``.

        >>> ser = pd.Series(pd.date_range('2000', periods=2, tz="CET"))
        >>> ser.to_numpy(dtype=object)
        array([Timestamp('2000-01-01 00:00:00+0100', tz='CET'),
               Timestamp('2000-01-02 00:00:00+0100', tz='CET')],
              dtype=object)

        Or ``dtype='datetime64[ns]'`` to return an ndarray of native
        datetime64 values. The values are converted to UTC and the timezone
        info is dropped.

        >>> ser.to_numpy(dtype="datetime64[ns]")
        ... # doctest: +ELLIPSIS
        array(['1999-12-31T23:00:00.000000000', '2000-01-01T23:00:00...'],
              dtype='datetime64[ns]')
        """
        ...
    
    @property
    def empty(self) -> bool:
        ...
    
    def max(self, axis=..., skipna: bool = ..., *args, **kwargs): # -> Dtype:
        """
        Return the maximum value of the Index.

        Parameters
        ----------
        axis : int, optional
            For compatibility with NumPy. Only 0 or None are allowed.
        skipna : bool, default True
            Exclude NA/null values when showing the result.
        *args, **kwargs
            Additional arguments and keywords for compatibility with NumPy.

        Returns
        -------
        scalar
            Maximum value.

        See Also
        --------
        Index.min : Return the minimum value in an Index.
        Series.max : Return the maximum value in a Series.
        DataFrame.max : Return the maximum values in a DataFrame.

        Examples
        --------
        >>> idx = pd.Index([3, 2, 1])
        >>> idx.max()
        3

        >>> idx = pd.Index(['c', 'b', 'a'])
        >>> idx.max()
        'c'

        For a MultiIndex, the maximum is determined lexicographically.

        >>> idx = pd.MultiIndex.from_product([('a', 'b'), (2, 1)])
        >>> idx.max()
        ('b', 2)
        """
        ...
    
    @doc(op="max", oppose="min", value="largest")
    def argmax(self, axis=..., skipna: bool = ..., *args, **kwargs) -> int:
        """
        Return int position of the {value} value in the Series.

        If the {op}imum is achieved in multiple locations,
        the first row position is returned.

        Parameters
        ----------
        axis : {{None}}
            Unused. Parameter needed for compatibility with DataFrame.
        skipna : bool, default True
            Exclude NA/null values when showing the result.
        *args, **kwargs
            Additional arguments and keywords for compatibility with NumPy.

        Returns
        -------
        int
            Row position of the {op}imum value.

        See Also
        --------
        Series.arg{op} : Return position of the {op}imum value.
        Series.arg{oppose} : Return position of the {oppose}imum value.
        numpy.ndarray.arg{op} : Equivalent method for numpy arrays.
        Series.idxmax : Return index label of the maximum values.
        Series.idxmin : Return index label of the minimum values.

        Examples
        --------
        Consider dataset containing cereal calories

        >>> s = pd.Series({{'Corn Flakes': 100.0, 'Almond Delight': 110.0,
        ...                'Cinnamon Toast Crunch': 120.0, 'Cocoa Puff': 110.0}})
        >>> s
        Corn Flakes              100.0
        Almond Delight           110.0
        Cinnamon Toast Crunch    120.0
        Cocoa Puff               110.0
        dtype: float64

        >>> s.argmax()
        2
        >>> s.argmin()
        0

        The maximum cereal calories is the third element and
        the minimum cereal calories is the first element,
        since series is zero-indexed.
        """
        ...
    
    def min(self, axis=..., skipna: bool = ..., *args, **kwargs): # -> Dtype:
        """
        Return the minimum value of the Index.

        Parameters
        ----------
        axis : {None}
            Dummy argument for consistency with Series.
        skipna : bool, default True
            Exclude NA/null values when showing the result.
        *args, **kwargs
            Additional arguments and keywords for compatibility with NumPy.

        Returns
        -------
        scalar
            Minimum value.

        See Also
        --------
        Index.max : Return the maximum value of the object.
        Series.min : Return the minimum value in a Series.
        DataFrame.min : Return the minimum values in a DataFrame.

        Examples
        --------
        >>> idx = pd.Index([3, 2, 1])
        >>> idx.min()
        1

        >>> idx = pd.Index(['c', 'b', 'a'])
        >>> idx.min()
        'a'

        For a MultiIndex, the minimum is determined lexicographically.

        >>> idx = pd.MultiIndex.from_product([('a', 'b'), (2, 1)])
        >>> idx.min()
        ('a', 1)
        """
        ...
    
    @doc(argmax, op="min", oppose="max", value="smallest")
    def argmin(self, axis=..., skipna=..., *args, **kwargs) -> int:
        ...
    
    def tolist(self): # -> list[Unknown] | Any:
        """
        Return a list of the values.

        These are each a scalar type, which is a Python scalar
        (for str, int, float) or a pandas scalar
        (for Timestamp/Timedelta/Interval/Period)

        Returns
        -------
        list

        See Also
        --------
        numpy.ndarray.tolist : Return the array as an a.ndim-levels deep
            nested list of Python scalars.
        """
        ...
    
    to_list = ...
    def __iter__(self): # -> Iterator[Any] | map[Unknown]:
        """
        Return an iterator of the values.

        These are each a scalar type, which is a Python scalar
        (for str, int, float) or a pandas scalar
        (for Timestamp/Timedelta/Interval/Period)

        Returns
        -------
        iterator
        """
        ...
    
    @cache_readonly
    def hasnans(self) -> bool:
        """
        Return True if there are any NaNs.

        Enables various performance speedups.
        """
        ...
    
    def isna(self) -> npt.NDArray[np.bool_]:
        ...
    
    def value_counts(self, normalize: bool = ..., sort: bool = ..., ascending: bool = ..., bins=..., dropna: bool = ...) -> Series:
        """
        Return a Series containing counts of unique values.

        The resulting object will be in descending order so that the
        first element is the most frequently-occurring element.
        Excludes NA values by default.

        Parameters
        ----------
        normalize : bool, default False
            If True then the object returned will contain the relative
            frequencies of the unique values.
        sort : bool, default True
            Sort by frequencies.
        ascending : bool, default False
            Sort in ascending order.
        bins : int, optional
            Rather than count values, group them into half-open bins,
            a convenience for ``pd.cut``, only works with numeric data.
        dropna : bool, default True
            Don't include counts of NaN.

        Returns
        -------
        Series

        See Also
        --------
        Series.count: Number of non-NA elements in a Series.
        DataFrame.count: Number of non-NA elements in a DataFrame.
        DataFrame.value_counts: Equivalent method on DataFrames.

        Examples
        --------
        >>> index = pd.Index([3, 1, 2, 3, 4, np.nan])
        >>> index.value_counts()
        3.0    2
        1.0    1
        2.0    1
        4.0    1
        dtype: int64

        With `normalize` set to `True`, returns the relative frequency by
        dividing all values by the sum of values.

        >>> s = pd.Series([3, 1, 2, 3, 4, np.nan])
        >>> s.value_counts(normalize=True)
        3.0    0.4
        1.0    0.2
        2.0    0.2
        4.0    0.2
        dtype: float64

        **bins**

        Bins can be useful for going from a continuous variable to a
        categorical variable; instead of counting unique
        apparitions of values, divide the index in the specified
        number of half-open bins.

        >>> s.value_counts(bins=3)
        (0.996, 2.0]    2
        (2.0, 3.0]      2
        (3.0, 4.0]      1
        dtype: int64

        **dropna**

        With `dropna` set to `False` we can also see NaN index values.

        >>> s.value_counts(dropna=False)
        3.0    2
        1.0    1
        2.0    1
        4.0    1
        NaN    1
        dtype: int64
        """
        ...
    
    def unique(self): # -> ArrayLike | NDArray[Any]:
        ...
    
    def nunique(self, dropna: bool = ...) -> int:
        """
        Return number of unique elements in the object.

        Excludes NA values by default.

        Parameters
        ----------
        dropna : bool, default True
            Don't include NaN in the count.

        Returns
        -------
        int

        See Also
        --------
        DataFrame.nunique: Method nunique for DataFrame.
        Series.count: Count non-NA/null observations in the Series.

        Examples
        --------
        >>> s = pd.Series([1, 3, 5, 7, 7])
        >>> s
        0    1
        1    3
        2    5
        3    7
        4    7
        dtype: int64

        >>> s.nunique()
        4
        """
        ...
    
    @property
    def is_unique(self) -> bool:
        """
        Return boolean if values in the object are unique.

        Returns
        -------
        bool
        """
        ...
    
    @property
    def is_monotonic(self) -> bool:
        """
        Return boolean if values in the object are monotonically increasing.

        .. deprecated:: 1.5.0
            is_monotonic is deprecated and will be removed in a future version.
            Use is_monotonic_increasing instead.

        Returns
        -------
        bool
        """
        ...
    
    @property
    def is_monotonic_increasing(self) -> bool:
        """
        Return boolean if values in the object are monotonically increasing.

        Returns
        -------
        bool
        """
        ...
    
    @property
    def is_monotonic_decreasing(self) -> bool:
        """
        Return boolean if values in the object are monotonically decreasing.

        Returns
        -------
        bool
        """
        ...
    
    @doc(algorithms.factorize, values="", order="", size_hint="", sort=textwrap.dedent("""\
            sort : bool, default False
                Sort `uniques` and shuffle `codes` to maintain the
                relationship.
            """))
    def factorize(self, sort: bool = ..., na_sentinel: int | lib.NoDefault = ..., use_na_sentinel: bool | lib.NoDefault = ...): # -> tuple[ndarray[Unknown, Unknown], ndarray[Unknown, Unknown] | Index]:
        ...
    
    @overload
    def searchsorted(self, value: npt._ScalarLike_co, side: Literal["left", "right"] = ..., sorter: NumpySorter = ...) -> np.intp:
        ...
    
    @overload
    def searchsorted(self, value: npt.ArrayLike | ExtensionArray, side: Literal["left", "right"] = ..., sorter: NumpySorter = ...) -> npt.NDArray[np.intp]:
        ...
    
    @doc(_shared_docs["searchsorted"], klass="Index")
    def searchsorted(self, value: NumpyValueArrayLike | ExtensionArray, side: Literal["left", "right"] = ..., sorter: NumpySorter = ...) -> npt.NDArray[np.intp] | np.intp:
        ...
    
    def drop_duplicates(self, keep=...):
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Any, TYPE_CHECKING
from pandas._typing import AnyArrayLike
from pandas.core.frame import DataFrame
from pandas.core.indexes.base import Index

"""
Low-dependency indexing utilities.
"""
if TYPE_CHECKING:
    ...
def is_valid_positional_slice(slc: slice) -> bool:
    """
    Check if a slice object can be interpreted as a positional indexer.

    Parameters
    ----------
    slc : slice

    Returns
    -------
    bool

    Notes
    -----
    A valid positional slice may also be interpreted as a label-based slice
    depending on the index being sliced.
    """
    ...

def is_list_like_indexer(key) -> bool:
    """
    Check if we have a list-like indexer that is *not* a NamedTuple.

    Parameters
    ----------
    key : object

    Returns
    -------
    bool
    """
    ...

def is_scalar_indexer(indexer, ndim: int) -> bool:
    """
    Return True if we are all scalar indexers.

    Parameters
    ----------
    indexer : object
    ndim : int
        Number of dimensions in the object being indexed.

    Returns
    -------
    bool
    """
    ...

def is_empty_indexer(indexer) -> bool:
    """
    Check if we have an empty indexer.

    Parameters
    ----------
    indexer : object

    Returns
    -------
    bool
    """
    ...

def check_setitem_lengths(indexer, value, values) -> bool:
    """
    Validate that value and indexer are the same length.

    An special-case is allowed for when the indexer is a boolean array
    and the number of true values equals the length of ``value``. In
    this case, no exception is raised.

    Parameters
    ----------
    indexer : sequence
        Key for the setitem.
    value : array-like
        Value for the setitem.
    values : array-like
        Values being set into.

    Returns
    -------
    bool
        Whether this is an empty listlike setting which is a no-op.

    Raises
    ------
    ValueError
        When the indexer is an ndarray or list and the lengths don't match.
    """
    ...

def validate_indices(indices: np.ndarray, n: int) -> None:
    """
    Perform bounds-checking for an indexer.

    -1 is allowed for indicating missing values.

    Parameters
    ----------
    indices : ndarray
    n : int
        Length of the array being indexed.

    Raises
    ------
    ValueError

    Examples
    --------
    >>> validate_indices(np.array([1, 2]), 3) # OK

    >>> validate_indices(np.array([1, -2]), 3)
    Traceback (most recent call last):
        ...
    ValueError: negative dimensions are not allowed

    >>> validate_indices(np.array([1, 2, 3]), 3)
    Traceback (most recent call last):
        ...
    IndexError: indices are out-of-bounds

    >>> validate_indices(np.array([-1, -1]), 0) # OK

    >>> validate_indices(np.array([0, 1]), 0)
    Traceback (most recent call last):
        ...
    IndexError: indices are out-of-bounds
    """
    ...

def maybe_convert_indices(indices, n: int, verify: bool = ...) -> np.ndarray:
    """
    Attempt to convert indices into valid, positive indices.

    If we have negative indices, translate to positive here.
    If we have indices that are out-of-bounds, raise an IndexError.

    Parameters
    ----------
    indices : array-like
        Array of indices that we are to convert.
    n : int
        Number of elements in the array that we are indexing.
    verify : bool, default True
        Check that all entries are between 0 and n - 1, inclusive.

    Returns
    -------
    array-like
        An array-like of positive indices that correspond to the ones
        that were passed in initially to this function.

    Raises
    ------
    IndexError
        One of the converted indices either exceeded the number of,
        elements (specified by `n`), or was still negative.
    """
    ...

def length_of_indexer(indexer, target=...) -> int:
    """
    Return the expected length of target[indexer]

    Returns
    -------
    int
    """
    ...

def deprecate_ndim_indexing(result, stacklevel: int = ...) -> None:
    """
    Helper function to raise the deprecation warning for multi-dimensional
    indexing on 1D Series/Index.

    GH#27125 indexer like idx[:, None] expands dim, but we cannot do that
    and keep an index, so we currently return ndarray, which is deprecated
    (Deprecation GH#30588).
    """
    ...

def unpack_1tuple(tup):
    """
    If we have a length-1 tuple/list that contains a slice, unpack to just
    the slice.

    Notes
    -----
    The list case is deprecated.
    """
    ...

def check_key_length(columns: Index, key, value: DataFrame) -> None:
    """
    Checks if a key used as indexer has the same length as the columns it is
    associated with.

    Parameters
    ----------
    columns : Index The columns of the DataFrame to index.
    key : A list-like of keys to index with.
    value : DataFrame The value to set for the keys.

    Raises
    ------
    ValueError: If the length of key is not equal to the number of columns in value
                or if the number of columns referenced by key is not equal to number
                of columns.
    """
    ...

def unpack_tuple_and_ellipses(item: tuple): # -> tuple[Unknown, ...]:
    """
    Possibly unpack arr[..., n] to arr[n]
    """
    ...

def check_array_indexer(array: AnyArrayLike, indexer: Any) -> Any:
    """
    Check if `indexer` is a valid array indexer for `array`.

    For a boolean mask, `array` and `indexer` are checked to have the same
    length. The dtype is validated, and if it is an integer or boolean
    ExtensionArray, it is checked if there are missing values present, and
    it is converted to the appropriate numpy array. Other dtypes will raise
    an error.

    Non-array indexers (integer, slice, Ellipsis, tuples, ..) are passed
    through as is.

    .. versionadded:: 1.0.0

    Parameters
    ----------
    array : array-like
        The array that is being indexed (only used for the length).
    indexer : array-like or list-like
        The array-like that's used to index. List-like input that is not yet
        a numpy array or an ExtensionArray is converted to one. Other input
        types are passed through as is.

    Returns
    -------
    numpy.ndarray
        The validated indexer as a numpy array that can be used to index.

    Raises
    ------
    IndexError
        When the lengths don't match.
    ValueError
        When `indexer` cannot be converted to a numpy ndarray to index
        (e.g. presence of missing values).

    See Also
    --------
    api.types.is_bool_dtype : Check if `key` is of boolean dtype.

    Examples
    --------
    When checking a boolean mask, a boolean ndarray is returned when the
    arguments are all valid.

    >>> mask = pd.array([True, False])
    >>> arr = pd.array([1, 2])
    >>> pd.api.indexers.check_array_indexer(arr, mask)
    array([ True, False])

    An IndexError is raised when the lengths don't match.

    >>> mask = pd.array([True, False, True])
    >>> pd.api.indexers.check_array_indexer(arr, mask)
    Traceback (most recent call last):
    ...
    IndexError: Boolean index has wrong length: 3 instead of 2.

    NA values in a boolean array are treated as False.

    >>> mask = pd.array([True, pd.NA])
    >>> pd.api.indexers.check_array_indexer(arr, mask)
    array([ True, False])

    A numpy boolean mask will get passed through (if the length is correct):

    >>> mask = np.array([True, False])
    >>> pd.api.indexers.check_array_indexer(arr, mask)
    array([ True, False])

    Similarly for integer indexers, an integer ndarray is returned when it is
    a valid indexer, otherwise an error is  (for integer indexers, a matching
    length is not required):

    >>> indexer = pd.array([0, 2], dtype="Int64")
    >>> arr = pd.array([1, 2, 3])
    >>> pd.api.indexers.check_array_indexer(arr, indexer)
    array([0, 2])

    >>> indexer = pd.array([0, pd.NA], dtype="Int64")
    >>> pd.api.indexers.check_array_indexer(arr, indexer)
    Traceback (most recent call last):
    ...
    ValueError: Cannot index with an integer indexer containing NA values

    For non-integer/boolean dtypes, an appropriate error is raised:

    >>> indexer = np.array([0., 2.], dtype="float64")
    >>> pd.api.indexers.check_array_indexer(arr, indexer)
    Traceback (most recent call last):
    ...
    IndexError: arrays used as indices must be of integer or boolean type
    """
    ...




"""
This type stub file was generated by pyright.
"""

import textwrap
import numpy as np
from typing import Any, Hashable
from pandas._typing import Dtype, IntervalClosedType, npt
from pandas.util._decorators import Appender, cache_readonly
from pandas.core.arrays.interval import IntervalArray, _interval_shared_docs
from pandas.core.indexes.base import Index, _index_shared_docs
from pandas.core.indexes.extension import ExtensionIndex, inherit_names

""" define the IntervalIndex """
_index_doc_kwargs = ...
@Appender(_interval_shared_docs["class"] % { "klass": "IntervalIndex","summary": "Immutable index of intervals that are closed on the same side.","name": _index_doc_kwargs["name"],"versionadded": "0.20.0","extra_attributes": "is_overlapping\nvalues\n","extra_methods": "","examples": textwrap.dedent("""\
    Examples
    --------
    A new ``IntervalIndex`` is typically constructed using
    :func:`interval_range`:

    >>> pd.interval_range(start=0, end=5)
    IntervalIndex([(0, 1], (1, 2], (2, 3], (3, 4], (4, 5]],
                  dtype='interval[int64, right]')

    It may also be constructed using one of the constructor
    methods: :meth:`IntervalIndex.from_arrays`,
    :meth:`IntervalIndex.from_breaks`, and :meth:`IntervalIndex.from_tuples`.

    See further examples in the doc strings of ``interval_range`` and the
    mentioned constructor methods.
    """) })
@inherit_names(["set_closed", "to_tuples"], IntervalArray, wrap=True)
@inherit_names(["__array__", "overlaps", "contains", "closed_left", "closed_right", "open_left", "open_right", "is_empty"], IntervalArray)
@inherit_names(["is_non_overlapping_monotonic", "closed"], IntervalArray, cache=True)
class IntervalIndex(ExtensionIndex):
    _typ = ...
    closed: IntervalClosedType
    is_non_overlapping_monotonic: bool
    closed_left: bool
    closed_right: bool
    open_left: bool
    open_right: bool
    _data: IntervalArray
    _values: IntervalArray
    _can_hold_strings = ...
    _data_cls = IntervalArray
    def __new__(cls, data, closed=..., dtype: Dtype | None = ..., copy: bool = ..., name: Hashable = ..., verify_integrity: bool = ...) -> IntervalIndex:
        ...
    
    @classmethod
    @Appender(_interval_shared_docs["from_breaks"] % { "klass": "IntervalIndex","examples": textwrap.dedent("""\
        Examples
        --------
        >>> pd.IntervalIndex.from_breaks([0, 1, 2, 3])
        IntervalIndex([(0, 1], (1, 2], (2, 3]],
                      dtype='interval[int64, right]')
        """) })
    def from_breaks(cls, breaks, closed: IntervalClosedType | None = ..., name: Hashable = ..., copy: bool = ..., dtype: Dtype | None = ...) -> IntervalIndex:
        ...
    
    @classmethod
    @Appender(_interval_shared_docs["from_arrays"] % { "klass": "IntervalIndex","examples": textwrap.dedent("""\
        Examples
        --------
        >>> pd.IntervalIndex.from_arrays([0, 1, 2], [1, 2, 3])
        IntervalIndex([(0, 1], (1, 2], (2, 3]],
                      dtype='interval[int64, right]')
        """) })
    def from_arrays(cls, left, right, closed: IntervalClosedType = ..., name: Hashable = ..., copy: bool = ..., dtype: Dtype | None = ...) -> IntervalIndex:
        ...
    
    @classmethod
    @Appender(_interval_shared_docs["from_tuples"] % { "klass": "IntervalIndex","examples": textwrap.dedent("""\
        Examples
        --------
        >>> pd.IntervalIndex.from_tuples([(0, 1), (1, 2)])
        IntervalIndex([(0, 1], (1, 2]],
                       dtype='interval[int64, right]')
        """) })
    def from_tuples(cls, data, closed: str = ..., name: Hashable = ..., copy: bool = ..., dtype: Dtype | None = ...) -> IntervalIndex:
        ...
    
    def __contains__(self, key: Any) -> bool:
        """
        return a boolean if this key is IN the index
        We *only* accept an Interval

        Parameters
        ----------
        key : Interval

        Returns
        -------
        bool
        """
        ...
    
    def __reduce__(self): # -> tuple[(cls: Unknown, d: Unknown) -> Unknown, tuple[Type[Self@IntervalIndex], dict[str, Unknown]], None]:
        ...
    
    @property
    def inferred_type(self) -> str:
        """Return a string of the type inferred from the values"""
        ...
    
    @Appender(Index.memory_usage.__doc__)
    def memory_usage(self, deep: bool = ...) -> int:
        ...
    
    @cache_readonly
    def is_monotonic_decreasing(self) -> bool:
        """
        Return True if the IntervalIndex is monotonic decreasing (only equal or
        decreasing values), else False
        """
        ...
    
    @cache_readonly
    def is_unique(self) -> bool:
        """
        Return True if the IntervalIndex contains unique elements, else False.
        """
        ...
    
    @property
    def is_overlapping(self) -> bool:
        """
        Return True if the IntervalIndex has overlapping intervals, else False.

        Two intervals overlap if they share a common point, including closed
        endpoints. Intervals that only have an open endpoint in common do not
        overlap.

        Returns
        -------
        bool
            Boolean indicating if the IntervalIndex has overlapping intervals.

        See Also
        --------
        Interval.overlaps : Check whether two Interval objects overlap.
        IntervalIndex.overlaps : Check an IntervalIndex elementwise for
            overlaps.

        Examples
        --------
        >>> index = pd.IntervalIndex.from_tuples([(0, 2), (1, 3), (4, 5)])
        >>> index
        IntervalIndex([(0, 2], (1, 3], (4, 5]],
              dtype='interval[int64, right]')
        >>> index.is_overlapping
        True

        Intervals that share closed endpoints overlap:

        >>> index = pd.interval_range(0, 3, closed='both')
        >>> index
        IntervalIndex([[0, 1], [1, 2], [2, 3]],
              dtype='interval[int64, both]')
        >>> index.is_overlapping
        True

        Intervals that only have an open endpoint in common do not overlap:

        >>> index = pd.interval_range(0, 3, closed='left')
        >>> index
        IntervalIndex([[0, 1), [1, 2), [2, 3)],
              dtype='interval[int64, left]')
        >>> index.is_overlapping
        False
        """
        ...
    
    def get_loc(self, key, method: str | None = ..., tolerance=...) -> int | slice | np.ndarray:
        """
        Get integer location, slice or boolean mask for requested label.

        Parameters
        ----------
        key : label
        method : {None}, optional
            * default: matches where the label is within an interval only.

        Returns
        -------
        int if unique index, slice if monotonic index, else mask

        Examples
        --------
        >>> i1, i2 = pd.Interval(0, 1), pd.Interval(1, 2)
        >>> index = pd.IntervalIndex([i1, i2])
        >>> index.get_loc(1)
        0

        You can also supply a point inside an interval.

        >>> index.get_loc(1.5)
        1

        If a label is in several intervals, you get the locations of all the
        relevant intervals.

        >>> i3 = pd.Interval(0, 2)
        >>> overlapping_index = pd.IntervalIndex([i1, i2, i3])
        >>> overlapping_index.get_loc(0.5)
        array([ True, False,  True])

        Only exact matches will be returned if an interval is provided.

        >>> index.get_loc(pd.Interval(0, 1))
        0
        """
        ...
    
    @Appender(_index_shared_docs["get_indexer_non_unique"] % _index_doc_kwargs)
    def get_indexer_non_unique(self, target: Index) -> tuple[npt.NDArray[np.intp], npt.NDArray[np.intp]]:
        ...
    
    _requires_unique_msg = ...
    @cache_readonly
    def left(self) -> Index:
        ...
    
    @cache_readonly
    def right(self) -> Index:
        ...
    
    @cache_readonly
    def mid(self) -> Index:
        ...
    
    @property
    def length(self) -> Index:
        ...
    


def interval_range(start=..., end=..., periods=..., freq=..., name: Hashable = ..., closed: IntervalClosedType = ...) -> IntervalIndex:
    """
    Return a fixed frequency IntervalIndex.

    Parameters
    ----------
    start : numeric or datetime-like, default None
        Left bound for generating intervals.
    end : numeric or datetime-like, default None
        Right bound for generating intervals.
    periods : int, default None
        Number of periods to generate.
    freq : numeric, str, or DateOffset, default None
        The length of each interval. Must be consistent with the type of start
        and end, e.g. 2 for numeric, or '5H' for datetime-like.  Default is 1
        for numeric and 'D' for datetime-like.
    name : str, default None
        Name of the resulting IntervalIndex.
    closed : {'left', 'right', 'both', 'neither'}, default 'right'
        Whether the intervals are closed on the left-side, right-side, both
        or neither.

    Returns
    -------
    IntervalIndex

    See Also
    --------
    IntervalIndex : An Index of intervals that are all closed on the same side.

    Notes
    -----
    Of the four parameters ``start``, ``end``, ``periods``, and ``freq``,
    exactly three must be specified. If ``freq`` is omitted, the resulting
    ``IntervalIndex`` will have ``periods`` linearly spaced elements between
    ``start`` and ``end``, inclusively.

    To learn more about datetime-like frequency strings, please see `this link
    <https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases>`__.

    Examples
    --------
    Numeric ``start`` and  ``end`` is supported.

    >>> pd.interval_range(start=0, end=5)
    IntervalIndex([(0, 1], (1, 2], (2, 3], (3, 4], (4, 5]],
                  dtype='interval[int64, right]')

    Additionally, datetime-like input is also supported.

    >>> pd.interval_range(start=pd.Timestamp('2017-01-01'),
    ...                   end=pd.Timestamp('2017-01-04'))
    IntervalIndex([(2017-01-01, 2017-01-02], (2017-01-02, 2017-01-03],
                   (2017-01-03, 2017-01-04]],
                  dtype='interval[datetime64[ns], right]')

    The ``freq`` parameter specifies the frequency between the left and right.
    endpoints of the individual intervals within the ``IntervalIndex``.  For
    numeric ``start`` and ``end``, the frequency must also be numeric.

    >>> pd.interval_range(start=0, periods=4, freq=1.5)
    IntervalIndex([(0.0, 1.5], (1.5, 3.0], (3.0, 4.5], (4.5, 6.0]],
                  dtype='interval[float64, right]')

    Similarly, for datetime-like ``start`` and ``end``, the frequency must be
    convertible to a DateOffset.

    >>> pd.interval_range(start=pd.Timestamp('2017-01-01'),
    ...                   periods=3, freq='MS')
    IntervalIndex([(2017-01-01, 2017-02-01], (2017-02-01, 2017-03-01],
                   (2017-03-01, 2017-04-01]],
                  dtype='interval[datetime64[ns], right]')

    Specify ``start``, ``end``, and ``periods``; the frequency is generated
    automatically (linearly spaced).

    >>> pd.interval_range(start=0, end=6, periods=4)
    IntervalIndex([(0.0, 1.5], (1.5, 3.0], (3.0, 4.5], (4.5, 6.0]],
              dtype='interval[float64, right]')

    The ``closed`` parameter specifies which endpoints of the individual
    intervals within the ``IntervalIndex`` are closed.

    >>> pd.interval_range(end=5, periods=4, closed='both')
    IntervalIndex([[1, 2], [2, 3], [3, 4], [4, 5]],
                  dtype='interval[int64, both]')
    """
    ...




# BoARIO : The Adaptative Regional Input Output model in python.
# Copyright (C) 2022  Samuel Juhel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations
from typing import Callable, Optional
from numpy.typing import ArrayLike
import warnings
import numpy as np
from boario import logger
import math
import inspect
from functools import partial

__all__ = ['Event']

def linear_recovery(elapsed_temporal_unit:int,init_kapital_destroyed:np.ndarray,recovery_time:int):
    """Linear Kapital recovery function

    Kapital is entirely recovered when `recovery_time` has passed since event
    started recovering

    Parameters
    ----------
    init_kapital_destroyed : float
        Initial kapital destroyed
    elapsed_temporal_unit : int
        Elapsed time since event started recovering
    recovery_time : int
        Total time it takes the event to fully recover

    Examples
    --------
    FIXME: Add docs.

    """

    return init_kapital_destroyed * (1-(elapsed_temporal_unit/recovery_time))

def convexe_recovery(elapsed_temporal_unit:int,init_kapital_destroyed:np.ndarray,recovery_time:int):
    """Convexe Kapital recovery function

    Kapital is recovered with characteristic time `recovery_time`. (This doesn't mean Kapital is fully recovered after this time !)
    This function models a recovery similar as the one happening in the rebuilding case, for the same characteristic time.

    Parameters
    ----------
    init_kapital_destroyed : float
        Initial kapital destroyed
    elapsed_temporal_unit : int
        Elapsed time since event started recovering
    recovery_time : int
        Total time it takes the event to fully recover

    """
    return init_kapital_destroyed * (1-(1/recovery_time))**elapsed_temporal_unit

def convexe_recovery_scaled(elapsed_temporal_unit:int,init_kapital_destroyed:np.ndarray,recovery_time:int,scaling_factor:float=4):
    """Convexe Kapital recovery function (scaled to match other recovery duration)

    Kapital is mostly recovered (>95% by default for most cases) when `recovery_time` has passed since event
    started recovering.

    Parameters
    ----------
    init_kapital_destroyed : float
        Initial kapital destroyed
    elapsed_temporal_unit : int
        Elapsed time since event started recovering
    recovery_time : int
        Total time it takes the event to fully recover
    scaling_factor: float
        Used to scale the exponent in the function so that kapital is mostly rebuilt after `recovery_time`.
    A value of 4 insure >95% of kapital is recovered for a reasonable range of `recovery_time` values.

    """
    return init_kapital_destroyed * (1-(elapsed_temporal_unit/recovery_time))**(scaling_factor*elapsed_temporal_unit)

def concave_recovery(elapsed_temporal_unit:int,init_kapital_destroyed:np.ndarray,recovery_time:int,steep_factor:float=0.000001,half_recovery_time:Optional[int]=None):
    """Concave (s-shaped) Kapital recovery function

    Kapital is mostly (>95% in most cases) recovered when `recovery_time` has passed since event started recovering.

    Parameters
    ----------
    init_kapital_destroyed : float
        Initial kapital destroyed
    elapsed_temporal_unit : int
        Elapsed time since event started recovering
    recovery_time : int
        Total time it takes the event to fully recover
    steep_factor: float
        This coefficient governs the slope of the central part of the s-shape, smaller values lead to a steeper slope. As such it also affect the percentage of kapital rebuilt
    after `recovery_time` has elapsed. A value of 0.000001 should insure 95% of the kapital is rebuild for a reasonable range of recovery duration.
    half_recovery_time : int
        This can by use to change the time the inflexion point of the s-shape curve is attained. By default it is set to half the recovery duration.

    """
    if half_recovery_time is None:
        tau_h = 2
    else:
        tau_h = recovery_time/half_recovery_time
    exponent = (np.log(recovery_time)-np.log(steep_factor))/(np.log(recovery_time)-np.log(tau_h))
    return (init_kapital_destroyed * recovery_time)/(recovery_time + steep_factor*(elapsed_temporal_unit**exponent))
class Event(object):
    possible_sectors : np.ndarray = None #type: ignore
    possible_regions : np.ndarray = None #type: ignore
    temporal_unit_range : int = None #type: ignore
    z_shape : tuple[int,int] = None #type: ignore
    y_shape : tuple[int,int] = None #type: ignore
    x_shape : tuple[int,int] = None #type: ignore
    regions_idx : np.ndarray = None #type: ignore
    sectors_idx : np.ndarray = None #type: ignore
    monetary_unit : int = None #type: ignore
    sectors_gva_shares : np.ndarray = None #type: ignore
    Z_distrib: np.ndarray = None #type: ignore
    mrio_name: str = ""

    def __init__(self, event:dict) -> None:
        super().__init__()
        if event.get("globals_var") is not None:
            for k,v in event["globals_var"].items():
                if Event.__dict__[k] != v:
                    logger.warning("""You are trying to load an event which was simulated under different global vars, this might break: key:{} with value {} in dict and {} in Event class""".format(k,v,Event.__dict__[k]))
        for k,v in Event.__dict__.items():
            if k!="__doc__" and v is None:
                raise AttributeError("Event Class attribute {} is not set yet so instantiating an Event isn't possible".format(k))
        self.shock_type = event["shock_type"]
        self.name = event.get("name","unnamed")
        self.occurrence = event.get("occur",1)
        self.duration = event.get("duration",1)
        self.aff_regions = event["aff_regions"]
        self.aff_sectors = event["aff_sectors"]
        self.dmg_regional_distrib = event.get("dmg_regional_distrib",[1/self.aff_regions.size for _ in range(self.aff_regions.size)])
        self.dmg_sectoral_distrib_type = event.get("dmg_sectoral_distrib_type","equally shared")
        self.dmg_sectoral_distrib = event.get("dmg_sectoral_distrib",[1/self.aff_sectors.size for _ in range(self.aff_sectors.size)])
        self.rebuilding_sectors = event.get("rebuilding_sectors")
        self.total_kapital_destroyed = event.get("kapital_damage")
        self.regional_sectoral_kapital_destroyed = None
        self.rebuilding_demand_house = None
        self.rebuilding_demand_indus = None
        self.production_share_allocated = None
        self.prod_cap_delta_arbitrary = None
        self.happened = False
        self._recoverable_kind : bool = False
        self._recoverable : bool = False
        self._rebuildable_kind : bool = False
        self._rebuildable : bool = False
        self._recovery_fun : Optional[Callable] = None
        self.rebuild_tau = event.get("rebuild_tau")
        self.over = False
        self.init_shock(event)
        self.event_dict = event
        self.event_dict["globals_vars"] = {
            "possible_sectors" : list(self.possible_sectors),
            "possible_regions" : list(self.possible_regions),
            "temporal_unit_range" : self.temporal_unit_range,
            "z_shape" : self.z_shape,
            "y_shape" : self.y_shape,
            "x_shape" : self.x_shape,
            "monetary_unit" : self.monetary_unit,
            "mrio_used" : self.mrio_name
            }

    @property
    def recovery_function(self)->Optional[Callable]:
        return self._recovery_fun

    @recovery_function.setter
    def recovery_function(self,r_fun:str|Callable|None):
        if self.shock_type == "kapital_destroyed_recover":
            if r_fun is None:
                r_fun = "linear"
            if self.recovery_time is None:
                raise AttributeError("Impossible to set recovery function if no recovery time is given.")
            if isinstance(r_fun,str):
                if r_fun == "linear":
                    fun = linear_recovery
                elif r_fun == "convexe":
                    fun = convexe_recovery_scaled
                elif r_fun == "convexe noscale":
                    fun = convexe_recovery
                elif r_fun == "concave":
                    fun = concave_recovery
                else:
                    raise NotImplementedError("No implemented recovery function corresponding to {}".format(r_fun))
            elif callable(r_fun):
                r_fun_argsspec = inspect.getfullargspec(r_fun)
                r_fun_args = r_fun_argsspec.args + r_fun_argsspec.kwonlyargs
                if not all(args in r_fun_args for args in ["init_kapital_destroyed","elapsed_temporal_unit","recovery_time"]):
                    raise ValueError("Recovery function has to have at least the following keyword arguments: {}".format(["init_kapital_destroyed","elapsed_temporal_unit","recovery_time"]))
                fun = r_fun

            else:
                raise ValueError("Given recovery function is not a str or callable")

            r_fun_partial = partial(fun, init_kapital_destroyed=self._regional_sectoral_kapital_destroyed_0, recovery_time=self.recovery_time)
            self._recovery_fun = r_fun_partial
        else:
            self._recovery_fun = None

    @property
    def occurrence(self)->int:
        return self._occur

    @occurrence.setter
    def occurrence(self, value:int):
        if not 0 <= value <= self.temporal_unit_range:
            raise ValueError("Occurrence of event is not in the range of simulation steps : {} not in [0-{}]".format(value, self.temporal_unit_range))
        else:
            self._occur=value

    @property
    def duration(self)->int:
        return self._duration

    @duration.setter
    def duration(self, value:int):
        if not 0 <= self.occurrence + value <= self.temporal_unit_range:
            raise ValueError("Occurrence + duration of event is not in the range of simulation steps : {} + {} not in [0-{}]".format(self.occurrence, value, self.temporal_unit_range))
        else:
            self._duration=value

    @property
    def aff_regions(self)->np.ndarray:
        return self._aff_regions

    @property
    def aff_regions_idx(self)->np.ndarray:
        return self._aff_regions_idx

    @aff_regions.setter
    def aff_regions(self, value:ArrayLike|str):
        if isinstance(value,str):
            value = [value]
        value = np.array(value)
        impossible_regions = np.setdiff1d(value,self.possible_regions)
        if impossible_regions.size > 0:
            raise ValueError("These regions are not in the model : {}".format(impossible_regions))
        else:
            self._aff_regions = value
            self._aff_regions_idx = np.searchsorted(self.possible_regions, value)

    @property
    def aff_sectors(self)->np.ndarray:
        return self._aff_sectors

    @property
    def aff_sectors_idx(self)->np.ndarray:
        return self._aff_sectors_idx

    @aff_sectors.setter
    def aff_sectors(self, value:ArrayLike|str):
        if isinstance(value,str):
            value = [value]
        value = np.array(value)
        impossible_sectors = np.setdiff1d(value,self.possible_sectors)
        if impossible_sectors.size > 0 :
            raise ValueError("These sectors are not in the model : {}".format(impossible_sectors))
        else:
            self._aff_sectors = value
            self._aff_sectors_idx = np.searchsorted(self.possible_sectors, value)

    @property
    def dmg_regional_distrib(self)->np.ndarray:
        return self._dmg_regional_distrib

    @dmg_regional_distrib.setter
    def dmg_regional_distrib(self,value:ArrayLike):
        if self.aff_regions is None:
            raise AttributeError("Affected regions attribute isn't set yet")
        value = np.array(value)
        if value.size != self.aff_regions.size:
            raise ValueError("There are {} affected regions by the event and length of given damage distribution is {}".format(self.aff_regions.size,value.size))
        s = value.sum()
        if not math.isclose(s,1):
            raise ValueError("Damage distribution doesn't sum up to 1 but to {}, which is not valid".format(s))
        self._dmg_regional_distrib = value

    @property
    def dmg_sectoral_distrib(self)->np.ndarray:
        return self._dmg_sectoral_distrib

    @dmg_sectoral_distrib.setter
    def dmg_sectoral_distrib(self,value:ArrayLike):
        if self.aff_sectors is None:
            raise AttributeError("Affected sectors attribute isn't set yet")
        value = np.array(value)
        if value.size != self.aff_sectors.size:
            raise ValueError("There are {} affected sectors by the event and length of given damage distribution is {}".format(self.aff_sectors.size,value.size))
        s = value.sum()
        if not math.isclose(s,1):
            raise ValueError("Damage distribution doesn't sum up to 1 but to {}, which is not valid".format(s))
        self._dmg_sectoral_distrib = value

    @property
    def dmg_sectoral_distrib_type(self)->str:
        return self._dmg_sectoral_distrib_type

    @dmg_sectoral_distrib_type.setter
    def dmg_sectoral_distrib_type(self,value:str):
        self._dmg_sectoral_distrib_type=value

    @property
    def rebuilding_sectors(self)->Optional[np.ndarray]:
        return self._rebuilding_sectors

    @property
    def rebuilding_sectors_idx(self)->Optional[np.ndarray]:
        return self._rebuilding_sectors_idx

    @property
    def rebuilding_sectors_shares(self)->Optional[np.ndarray]:
        return self._rebuilding_sectors_shares

    @rebuilding_sectors.setter
    def rebuilding_sectors(self, value:dict[str,float]|None):
        if value is None:
            self._rebuilding_sectors = None
            self._rebuilding_sectors_idx = None
        else:
            reb_sectors = np.array(list(value.keys()))
            reb_shares = np.array(list(value.values()))
            impossible_sectors = np.setdiff1d(reb_sectors,self.possible_sectors)
            if impossible_sectors.size > 0 :
                raise ValueError("These sectors are not in the model : {}".format(impossible_sectors))
            else:
                self._rebuilding_sectors = reb_sectors
                self._rebuilding_sectors_idx = np.searchsorted(self.possible_sectors, reb_sectors)
                self._rebuilding_sectors_shares = reb_shares

    @property
    def regional_sectoral_kapital_destroyed(self)->Optional[np.ndarray]:
        return self._regional_sectoral_kapital_destroyed

    @regional_sectoral_kapital_destroyed.setter
    def regional_sectoral_kapital_destroyed(self,value:ArrayLike|None):
        if value is None:
            self._regional_sectoral_kapital_destroyed = None
        else:
            value = np.array(value)
            if value.shape != self.x_shape:
                raise ValueError("Incorrect shape give for regional_sectoral_kapital_destroyed: {} given and {} expected".format(value.shape, self.x_shape))
            self._regional_sectoral_kapital_destroyed = value

    @property
    def rebuilding_demand_house(self)->Optional[np.ndarray]:
        return self._rebuilding_demand_house

    @rebuilding_demand_house.setter
    def rebuilding_demand_house(self,value:ArrayLike|None):
        if value is None:
            self._rebuilding_demand_house = None
        else:
            value = np.array(value)
            if value.shape != self.y_shape:
                raise ValueError("Incorrect shape give for rebuilding_demand_house: {} given and {} expected".format(value.shape, self.y_shape))
            self._rebuilding_demand_house = value

    @property
    def rebuilding_demand_indus(self)->Optional[np.ndarray]:
        return self._rebuilding_demand_indus

    @rebuilding_demand_indus.setter
    def rebuilding_demand_indus(self,value:ArrayLike|None):
        if value is None:
            self._rebuilding_demand_indus = value
        else:
            value = np.array(value)
            if value.shape != self.z_shape:
                raise ValueError("Incorrect shape give for rebuilding_demand_indus: {} given and {} expected".format(value.shape, self.z_shape))
            self._rebuilding_demand_indus = value
            # Also update kapital destroyed
            self._regional_sectoral_kapital_destroyed = value.sum(axis=0)

    @property
    def recoverable(self)->Optional[bool]:
        return self._recoverable

    @recoverable.setter
    def recoverable(self,current_temporal_unit:int):
        if self._recoverable_kind:
            reb = (self.occurrence + self.duration) <= current_temporal_unit
            if reb and not self.recoverable :
                logger.info("Temporal_Unit : {} ~ Event named {} that occured at {} in {} for {} damages has started recovering (no rebuilding demand)".format(current_temporal_unit,self.name,self.occurrence, self._aff_regions, self.total_kapital_destroyed))
            self._recoverable = reb

    @property
    def rebuildable(self)->Optional[bool]:
        return self._rebuildable

    @rebuildable.setter
    def rebuildable(self,current_temporal_unit:int):
        if self._rebuildable_kind:
            reb = (self.occurrence + self.duration) <= current_temporal_unit
            if reb and not self.rebuildable :
                logger.info("Temporal_Unit : {} ~ Event named {} that occured at {} in {} for {} damages has started rebuilding".format(current_temporal_unit,self.name,self.occurrence, self._aff_regions, self.total_kapital_destroyed))
            self._rebuildable = reb

    @property
    def prod_cap_delta_arbitrary(self)->Optional[np.ndarray]:
        return self._prod_cap_delta_arbitrary

    @prod_cap_delta_arbitrary.setter
    def prod_cap_delta_arbitrary(self, value:dict[str,float]|None):
        if value is None:
            self._prod_cap_delta_arbitrary = None
        else:
            if self.aff_regions is None:
                raise AttributeError("Affected regions attribute isn't set yet")
            aff_sectors = np.array(list(value.keys()))
            aff_shares = np.array(list(value.values()))
            impossible_sectors = np.setdiff1d(aff_sectors,self.possible_sectors)
            if impossible_sectors.size > 0 :
                raise ValueError("These sectors are not in the model : {}".format(impossible_sectors))
            self._aff_sectors = aff_sectors
            self._aff_sectors_idx = np.searchsorted(self.possible_sectors, aff_sectors)
            aff_industries_idx = np.array([self.possible_sectors.size * ri + si for ri in self.regions_idx for si in self._aff_sectors_idx])
            self._prod_cap_delta_arbitrary = np.zeros(shape=self.possible_sectors.size)
            self._prod_cap_delta_arbitrary[aff_industries_idx] = np.tile(aff_shares,self._aff_regions.size)

    def init_shock(self, event:dict):
        """Initiate shock from the event. Methods called by __init__.

        Sets the rebuilding demand for households and industry.

        First, if multiple regions are affected, it computes the vector of how damages are distributed across these.
        Then it computes the vector of how regional damages are distributed across affected sectors.
        It produces a ``n_regions`` * ``n_sectors`` sized vector hence stores the damage (i.e. capital destroyed) for all industries.

        This method also computes the `rebuilding demand` matrices from households and industries, i.e. the demand addressed to the rebuilding
        industries consequent to the shock.

        See :ref:`How to define Events <boario-events>` for further detail on how to parameter these distribution.

        Parameters
        ----------
        event_to_add_id : int
            The id (rank it the ``events`` list) of the event to shock the model with.

        Raises
        ------
        ValueError
            Raised if the production share allocated to rebuilding (in either
            the impacted regions or the others) is not in [0,1].
        """

        if self.shock_type in {"kapital_destroyed_rebuild","kapital_destroyed_recover"}:
            if self.total_kapital_destroyed is None:
                raise AttributeError("Total kapital destroyed isn't set yet, this shouldn't happen.")
            regions_idx = np.arange(self.possible_regions.size)
            aff_industries_idx = np.array([self.possible_sectors.size * ri + si for ri in self._aff_regions_idx for si in self._aff_sectors_idx])
            if not isinstance(self.total_kapital_destroyed, (int,float)):
                raise ValueError("Kapital damages is {}, which is not valid".format(type(self.total_kapital_destroyed)))
            self.total_kapital_destroyed /= self.monetary_unit
            self.remaining_kapital_destroyed = self.total_kapital_destroyed
            regional_damages = np.array(self.dmg_regional_distrib) * self.total_kapital_destroyed
            # GDP CASE
            if self.dmg_sectoral_distrib_type == "gdp":
                shares = self.sectors_gva_shares.reshape((self.possible_regions.size,self.possible_sectors.size))
                self.dmg_sectoral_distrib = (shares[self._aff_regions_idx][:,self._aff_sectors_idx]/shares[self._aff_regions_idx][:,self._aff_sectors_idx].sum(axis=1)[:,np.newaxis])

            regional_sectoral_damages = regional_damages * self.dmg_sectoral_distrib
            tmp = np.zeros(self.x_shape,dtype="float")
            tmp[aff_industries_idx] = regional_sectoral_damages
            self._regional_sectoral_kapital_destroyed_0 = tmp.copy()
            self.regional_sectoral_kapital_destroyed = tmp.copy()
            if self._regional_sectoral_kapital_destroyed is None:
                    raise ValueError("Rebuilding sectors are not set for this event")
            if self.shock_type == "kapital_destroyed_rebuild":
                self._rebuildable_kind = True
            # REBUILDING
                if self._rebuilding_sectors_idx is None:
                    raise ValueError("Rebuilding sectors are not set for this event")
                rebuilding_industries_idx = np.array([self.possible_sectors.size * ri + si for ri in self._aff_regions_idx for si in self._rebuilding_sectors_idx])
                rebuilding_industries_RoW_idx = np.array([self.possible_sectors.size * ri + si for ri in regions_idx if ri not in self._aff_regions_idx for si in self._rebuilding_sectors_idx])
                rebuilding_demand = np.outer(self._rebuilding_sectors_shares,regional_sectoral_damages)
                tmp = np.zeros(self.z_shape,dtype="float")
                mask = np.ix_(np.union1d(rebuilding_industries_RoW_idx,rebuilding_industries_idx),aff_industries_idx)

                tmp[mask] = self.Z_distrib[mask] * np.tile(rebuilding_demand, (self.possible_regions.size,1))
                self.rebuilding_demand_indus = tmp
                self.rebuilding_demand_house = np.zeros(shape=self.y_shape)
            else:
                self._rebuildable_kind = False
                self._recoverable_kind = True
                self.recovery_time = event["recovery_time"]
                self.recovery_function = event.get("recovery_function")



        elif self.shock_type == "production_capacity_loss":
            self.prod_cap_delta = event["prod_cap_delta"]

    def recovery(self,current_temporal_unit:int):
        if not self._recoverable_kind or self._rebuildable_kind:
            raise AttributeError("Event is not initiated as a recoverable event")
        elapsed = current_temporal_unit - (self.occurrence + self.duration)
        if elapsed < 0:
            raise RuntimeError("Trying to recover before event is over")
        if self.recovery_function is None:
            raise RuntimeError("Trying to recover event while recovery function isn't set yet")
        res = self.recovery_function(elapsed_temporal_unit=elapsed)
        precision = int(math.log10(self.monetary_unit)) + 1
        res = np.around(res,precision)
        self.regional_sectoral_kapital_destroyed = res

    def __repr__(self):
        #TODO: find ways to represent long lists
        return f''' [Representation WIP]
        Event(
              name = {self.name},
              occur = {self.occurrence},
              duration = {self.duration}
              aff_regions = {self.aff_regions},
              aff_sectors = {self.aff_sectors},
             )
        '''



"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Any, TYPE_CHECKING
from pandas._typing import ArrayLike, Axis, npt
from pandas import Index

"""
Routines for filling missing data.
"""
if TYPE_CHECKING:
    ...
def check_value_size(value, mask: npt.NDArray[np.bool_], length: int):
    """
    Validate the size of the values passed to ExtensionArray.fillna.
    """
    ...

def mask_missing(arr: ArrayLike, values_to_mask) -> npt.NDArray[np.bool_]:
    """
    Return a masking array of same size/shape as arr
    with entries equaling any member of values_to_mask set to True

    Parameters
    ----------
    arr : ArrayLike
    values_to_mask: list, tuple, or scalar

    Returns
    -------
    np.ndarray[bool]
    """
    ...

def clean_fill_method(method: str | None, allow_nearest: bool = ...): # -> str | None:
    ...

NP_METHODS = ...
SP_METHODS = ...
def clean_interp_method(method: str, index: Index, **kwargs) -> str:
    ...

def find_valid_index(values, *, how: str) -> int | None:
    """
    Retrieves the index of the first valid value.

    Parameters
    ----------
    values : ndarray or ExtensionArray
    how : {'first', 'last'}
        Use this parameter to change between the first or last valid index.

    Returns
    -------
    int or None
    """
    ...

def interpolate_array_2d(data: np.ndarray, method: str = ..., axis: int = ..., index: Index | None = ..., limit: int | None = ..., limit_direction: str = ..., limit_area: str | None = ..., fill_value: Any | None = ..., coerce: bool = ..., downcast: str | None = ..., **kwargs) -> None:
    """
    Wrapper to dispatch to either interpolate_2d or _interpolate_2d_with_fill.

    Notes
    -----
    Alters 'data' in-place.
    """
    ...

def interpolate_2d(values: np.ndarray, method: str = ..., axis: Axis = ..., limit: int | None = ..., limit_area: str | None = ...) -> None:
    """
    Perform an actual interpolation of values, values will be make 2-d if
    needed fills inplace, returns the result.

    Parameters
    ----------
    values: np.ndarray
        Input array.
    method: str, default "pad"
        Interpolation method. Could be "bfill" or "pad"
    axis: 0 or 1
        Interpolation axis
    limit: int, optional
        Index limit on interpolation.
    limit_area: str, optional
        Limit area for interpolation. Can be "inside" or "outside"

    Notes
    -----
    Modifies values in-place.
    """
    ...

_fill_methods = ...
def get_fill_func(method, ndim: int = ...): # -> (values: ndarray[Unknown, Unknown], limit: int | None = None, mask: NDArray[bool_] | None = None) -> tuple[ndarray[Unknown, Unknown], NDArray[bool_]]:
    ...

def clean_reindex_fill_method(method) -> str | None:
    ...




"""
This type stub file was generated by pyright.
"""

from pandas._typing import ArrayLike

""" basic inference routines """
is_bool = ...
is_integer = ...
is_float = ...
is_complex = ...
is_scalar = ...
is_decimal = ...
is_interval = ...
is_list_like = ...
is_iterator = ...
def is_number(obj) -> bool:
    """
    Check if the object is a number.

    Returns True when the object is a number, and False if is not.

    Parameters
    ----------
    obj : any type
        The object to check if is a number.

    Returns
    -------
    is_number : bool
        Whether `obj` is a number or not.

    See Also
    --------
    api.types.is_integer: Checks a subgroup of numbers.

    Examples
    --------
    >>> from pandas.api.types import is_number
    >>> is_number(1)
    True
    >>> is_number(7.15)
    True

    Booleans are valid because they are int subclass.

    >>> is_number(False)
    True

    >>> is_number("foo")
    False
    >>> is_number("5")
    False
    """
    ...

def iterable_not_string(obj) -> bool:
    """
    Check if the object is an iterable but not a string.

    Parameters
    ----------
    obj : The object to check.

    Returns
    -------
    is_iter_not_string : bool
        Whether `obj` is a non-string iterable.

    Examples
    --------
    >>> iterable_not_string([1, 2, 3])
    True
    >>> iterable_not_string("foo")
    False
    >>> iterable_not_string(1)
    False
    """
    ...

def is_file_like(obj) -> bool:
    """
    Check if the object is a file-like object.

    For objects to be considered file-like, they must
    be an iterator AND have either a `read` and/or `write`
    method as an attribute.

    Note: file-like objects must be iterable, but
    iterable objects need not be file-like.

    Parameters
    ----------
    obj : The object to check

    Returns
    -------
    is_file_like : bool
        Whether `obj` has file-like properties.

    Examples
    --------
    >>> import io
    >>> buffer = io.StringIO("data")
    >>> is_file_like(buffer)
    True
    >>> is_file_like([1, 2, 3])
    False
    """
    ...

def is_re(obj) -> bool:
    """
    Check if the object is a regex pattern instance.

    Parameters
    ----------
    obj : The object to check

    Returns
    -------
    is_regex : bool
        Whether `obj` is a regex pattern.

    Examples
    --------
    >>> is_re(re.compile(".*"))
    True
    >>> is_re("foo")
    False
    """
    ...

def is_re_compilable(obj) -> bool:
    """
    Check if the object can be compiled into a regex pattern instance.

    Parameters
    ----------
    obj : The object to check

    Returns
    -------
    is_regex_compilable : bool
        Whether `obj` can be compiled as a regex pattern.

    Examples
    --------
    >>> is_re_compilable(".*")
    True
    >>> is_re_compilable(1)
    False
    """
    ...

def is_array_like(obj) -> bool:
    """
    Check if the object is array-like.

    For an object to be considered array-like, it must be list-like and
    have a `dtype` attribute.

    Parameters
    ----------
    obj : The object to check

    Returns
    -------
    is_array_like : bool
        Whether `obj` has array-like properties.

    Examples
    --------
    >>> is_array_like(np.array([1, 2, 3]))
    True
    >>> is_array_like(pd.Series(["a", "b"]))
    True
    >>> is_array_like(pd.Index(["2016-01-01"]))
    True
    >>> is_array_like([1, 2, 3])
    False
    >>> is_array_like(("a", "b"))
    False
    """
    ...

def is_nested_list_like(obj) -> bool:
    """
    Check if the object is list-like, and that all of its elements
    are also list-like.

    Parameters
    ----------
    obj : The object to check

    Returns
    -------
    is_list_like : bool
        Whether `obj` has list-like properties.

    Examples
    --------
    >>> is_nested_list_like([[1, 2, 3]])
    True
    >>> is_nested_list_like([{1, 2, 3}, {1, 2, 3}])
    True
    >>> is_nested_list_like(["foo"])
    False
    >>> is_nested_list_like([])
    False
    >>> is_nested_list_like([[1, 2, 3], 1])
    False

    Notes
    -----
    This won't reliably detect whether a consumable iterator (e. g.
    a generator) is a nested-list-like without consuming the iterator.
    To avoid consuming it, we always return False if the outer container
    doesn't define `__len__`.

    See Also
    --------
    is_list_like
    """
    ...

def is_dict_like(obj) -> bool:
    """
    Check if the object is dict-like.

    Parameters
    ----------
    obj : The object to check

    Returns
    -------
    is_dict_like : bool
        Whether `obj` has dict-like properties.

    Examples
    --------
    >>> is_dict_like({1: 2})
    True
    >>> is_dict_like([1, 2, 3])
    False
    >>> is_dict_like(dict)
    False
    >>> is_dict_like(dict())
    True
    """
    ...

def is_named_tuple(obj) -> bool:
    """
    Check if the object is a named tuple.

    Parameters
    ----------
    obj : The object to check

    Returns
    -------
    is_named_tuple : bool
        Whether `obj` is a named tuple.

    Examples
    --------
    >>> from collections import namedtuple
    >>> Point = namedtuple("Point", ["x", "y"])
    >>> p = Point(1, 2)
    >>>
    >>> is_named_tuple(p)
    True
    >>> is_named_tuple((1, 2))
    False
    """
    ...

def is_hashable(obj) -> bool:
    """
    Return True if hash(obj) will succeed, False otherwise.

    Some types will pass a test against collections.abc.Hashable but fail when
    they are actually hashed with hash().

    Distinguish between these and other types by trying the call to hash() and
    seeing if they raise TypeError.

    Returns
    -------
    bool

    Examples
    --------
    >>> import collections
    >>> a = ([],)
    >>> isinstance(a, collections.abc.Hashable)
    True
    >>> is_hashable(a)
    False
    """
    ...

def is_sequence(obj) -> bool:
    """
    Check if the object is a sequence of objects.
    String types are not included as sequences here.

    Parameters
    ----------
    obj : The object to check

    Returns
    -------
    is_sequence : bool
        Whether `obj` is a sequence of objects.

    Examples
    --------
    >>> l = [1, 2, 3]
    >>>
    >>> is_sequence(l)
    True
    >>> is_sequence(iter(l))
    False
    """
    ...

def is_dataclass(item): # -> bool:
    """
    Checks if the object is a data-class instance

    Parameters
    ----------
    item : object

    Returns
    --------
    is_dataclass : bool
        True if the item is an instance of a data-class,
        will return false if you pass the data class itself

    Examples
    --------
    >>> from dataclasses import dataclass
    >>> @dataclass
    ... class Point:
    ...     x: int
    ...     y: int

    >>> is_dataclass(Point)
    False
    >>> is_dataclass(Point(0,2))
    True

    """
    ...

def is_inferred_bool_dtype(arr: ArrayLike) -> bool:
    """
    Check if this is a ndarray[bool] or an ndarray[object] of bool objects.

    Parameters
    ----------
    arr : np.ndarray or ExtensionArray

    Returns
    -------
    bool

    Notes
    -----
    This does not include the special treatment is_bool_dtype uses for
    Categorical.
    """
    ...




# BoARIO : The Adaptative Regional Input Output model in python.
# Copyright (C) 2022  Samuel Juhel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pathlib
import pymrio as pym
import pandas as pd
import pickle as pkl
import logging
import datetime
import argparse
import json
import country_converter as coco

parser = argparse.ArgumentParser(description='Aggregate an exio3 MRIO sectors')
parser.add_argument('exio_path', type=str, help='The str path to the exio3 zip file')
parser.add_argument('aggreg_path', type=str, help='The str path to the ods aggregation matrix file')
parser.add_argument('sector_names_json_path', type=str, help='The str path to the json file with the new names of the sectors')
parser.add_argument('-r', "--region_agg", type=str, help='The str path to the json file with the regions aggregation', nargs='?', default=None)
parser.add_argument('-o', "--output", type=str, help='The str path to save the pickled mrio to', nargs='?', default='./mrio_dump')

args = parser.parse_args()
logFormatter = logging.Formatter("%(asctime)s [%(levelname)-5.5s] %(name)s %(message)s", datefmt="%H:%M:%S")
scriptLogger = logging.getLogger("aggreg_exio3")
consoleHandler = logging.StreamHandler()
consoleHandler.setFormatter(logFormatter)

scriptLogger.addHandler(consoleHandler)
scriptLogger.setLevel(logging.INFO)
scriptLogger.propagate = False

agg_dicts_default = [{'aggregates': {'MX': 'MX'}, 'missing': 'Rest of the world'},
 {'aggregates': {'NO': 'NO'}, 'missing': 'Rest of the world'},
 {'aggregates': {'NL': 'NL'}, 'missing': 'Rest of the world'},
 {'aggregates': {'RU': 'RU'}, 'missing': 'Rest of the world'},
 {'aggregates': {'EE': 'EE'}, 'missing': 'Rest of the world'},
 {'aggregates': {'ID': 'ID'}, 'missing': 'Rest of the world'},
 {'aggregates': {'FR': 'FR'}, 'missing': 'Rest of the world'},
 {'aggregates': {'SE': 'SE'}, 'missing': 'Rest of the world'},
 {'aggregates': {'CN': 'CN'}, 'missing': 'Rest of the world'},
 {'aggregates': {'HU': 'HU'}, 'missing': 'Rest of the world'},
 {'aggregates': {'PT': 'PT'}, 'missing': 'Rest of the world'},
 {'aggregates': {'LT': 'LT'}, 'missing': 'Rest of the world'},
 {'aggregates': {'LV': 'LV'}, 'missing': 'Rest of the world'},
 {'aggregates': {'BR': 'BR'}, 'missing': 'Rest of the world'},
 {'aggregates': {'HR': 'HR'}, 'missing': 'Rest of the world'},
 {'aggregates': {'SK': 'SK'}, 'missing': 'Rest of the world'},
 {'aggregates': {'JP': 'JP'}, 'missing': 'Rest of the world'},
 {'aggregates': {'US': 'US'}, 'missing': 'Rest of the world'},
 {'aggregates': {'ZA': 'ZA'}, 'missing': 'Rest of the world'},
 {'aggregates': {'CH': 'CH'}, 'missing': 'Rest of the world'},
 {'aggregates': {'BE': 'BE'}, 'missing': 'Rest of the world'},
 {'aggregates': {'PL': 'PL'}, 'missing': 'Rest of the world'},
 {'aggregates': {'CA': 'CA'}, 'missing': 'Rest of the world'},
 {'aggregates': {'CZ': 'CZ'}, 'missing': 'Rest of the world'},
 {'aggregates': {'RO': 'RO'}, 'missing': 'Rest of the world'},
 {'aggregates': {'IT': 'IT'}, 'missing': 'Rest of the world'},
 {'aggregates': {'TR': 'TR'}, 'missing': 'Rest of the world'},
 {'aggregates': {'GB': 'GB'}, 'missing': 'Rest of the world'},
 {'aggregates': {'SI': 'SI'}, 'missing': 'Rest of the world'},
 {'aggregates': {'GR': 'GR'}, 'missing': 'Rest of the world'},
 {'aggregates': {'FI': 'FI'}, 'missing': 'Rest of the world'},
 {'aggregates': {'AU': 'AU'}, 'missing': 'Rest of the world'},
 {'aggregates': {'DE': 'DE'}, 'missing': 'Rest of the world'},
 {'aggregates': {'KR': 'KR'}, 'missing': 'Rest of the world'},
 {'aggregates': {'IE': 'IE'}, 'missing': 'Rest of the world'},
 {'aggregates': {'ES': 'ES'}, 'missing': 'Rest of the world'},
 {'aggregates': {'BG': 'BG'}, 'missing': 'Rest of the world'},
 {'aggregates': {'IN': 'IN'}, 'missing': 'Rest of the world'},
 {'aggregates': {'AT': 'AT'}, 'missing': 'Rest of the world'}]

def aggreg(exio_path, sector_aggregator_path, new_sectors_name_path, regions_aggregator_dicts, save_path=None):
    with pathlib.Path(new_sectors_name_path).open('r') as f:
        a = json.load(f)
    sec_agg_matrix = pd.read_excel(sector_aggregator_path, sheet_name="input", engine="odf", header=None).to_numpy()
    scriptLogger.info("Parsing exiobase3 from {}".format(pathlib.Path(exio_path).resolve()))
    exio3 = pym.parse_exiobase3(path=exio_path)
    scriptLogger.info("Done")
    scriptLogger.info("Computing the IO components")
    exio3.calc_all()
    scriptLogger.info("Done")
    scriptLogger.info("Reading aggregation matrix from sheet 'input' in file {}".format(pathlib.Path(sector_aggregator_path).absolute()))
    scriptLogger.info("Aggregating from {} to {} sectors".format(len(exio3.get_sectors()), sec_agg_matrix.shape[0]))
    exio3.aggregate(sector_agg=sec_agg_matrix)
    scriptLogger.info("Done")
    scriptLogger.info("Renaming sectors from {}".format(pathlib.Path(new_sectors_name_path).absolute()))
    exio3.rename_sectors(a)
    scriptLogger.info("Done")
    for dico in regions_aggregator_dicts:
        exio3_c = exio3.copy()
        regions_aggregator = coco.agg_conc(original_countries=exio3.get_regions(),
                                           aggregates=dico['aggregates'],
                                           missing_countries=dico['missing'])
        exio3_c.aggregate(region_agg=regions_aggregator)
        exio3_c.calc_all()
        name = save_path+list(dico['aggregates'].keys())[0]+"_"+datetime.datetime.now().strftime("%m-%d-%H%M")+".pkl"
        scriptLogger.info("Saving to {}".format(pathlib.Path(name).resolve()))
        with open(name, 'wb') as f:
            pkl.dump(exio3_c, f)

if __name__ == '__main__':
    args = parser.parse_args()
    name = pathlib.Path(args.exio_path).stem
    if args.region_agg is None:
        agg_dicts = agg_dicts_default
    else :
        with pathlib.Path(args.region_agg).open('r') as f:
            agg_dicts = json.load(f)
    aggreg(args.exio_path, args.aggreg_path, args.sector_names_json_path, regions_aggregator_dicts=agg_dicts, save_path=args.output)



"""
This type stub file was generated by pyright.
"""

import numpy as np
import pandas as pd
from typing import Any
from pandas.core.interchange.dataframe_protocol import Buffer, Column, DataFrame as DataFrameXchg, DtypeKind

_NP_DTYPES: dict[DtypeKind, dict[int, Any]] = ...
def from_dataframe(df, allow_copy=...) -> pd.DataFrame:
    """
    Build a ``pd.DataFrame`` from any DataFrame supporting the interchange protocol.

    Parameters
    ----------
    df : DataFrameXchg
        Object supporting the interchange protocol, i.e. `__dataframe__` method.
    allow_copy : bool, default: True
        Whether to allow copying the memory to perform the conversion
        (if false then zero-copy approach is requested).

    Returns
    -------
    pd.DataFrame
    """
    ...

def protocol_df_chunk_to_pandas(df: DataFrameXchg) -> pd.DataFrame:
    """
    Convert interchange protocol chunk to ``pd.DataFrame``.

    Parameters
    ----------
    df : DataFrameXchg

    Returns
    -------
    pd.DataFrame
    """
    ...

def primitive_column_to_ndarray(col: Column) -> tuple[np.ndarray, Any]:
    """
    Convert a column holding one of the primitive dtypes to a NumPy array.

    A primitive type is one of: int, uint, float, bool.

    Parameters
    ----------
    col : Column

    Returns
    -------
    tuple
        Tuple of np.ndarray holding the data and the memory owner object
        that keeps the memory alive.
    """
    ...

def categorical_column_to_series(col: Column) -> tuple[pd.Series, Any]:
    """
    Convert a column holding categorical data to a pandas Series.

    Parameters
    ----------
    col : Column

    Returns
    -------
    tuple
        Tuple of pd.Series holding the data and the memory owner object
        that keeps the memory alive.
    """
    ...

def string_column_to_ndarray(col: Column) -> tuple[np.ndarray, Any]:
    """
    Convert a column holding string data to a NumPy array.

    Parameters
    ----------
    col : Column

    Returns
    -------
    tuple
        Tuple of np.ndarray holding the data and the memory owner object
        that keeps the memory alive.
    """
    ...

def parse_datetime_format_str(format_str, data):
    """Parse datetime `format_str` to interpret the `data`."""
    ...

def datetime_column_to_ndarray(col: Column) -> tuple[np.ndarray, Any]:
    """
    Convert a column holding DateTime data to a NumPy array.

    Parameters
    ----------
    col : Column

    Returns
    -------
    tuple
        Tuple of np.ndarray holding the data and the memory owner object
        that keeps the memory alive.
    """
    ...

def buffer_to_ndarray(buffer: Buffer, dtype: tuple[DtypeKind, int, str, str], offset: int = ..., length: int | None = ...) -> np.ndarray:
    """
    Build a NumPy array from the passed buffer.

    Parameters
    ----------
    buffer : Buffer
        Buffer to build a NumPy array from.
    dtype : tuple
        Data type of the buffer conforming protocol dtypes format.
    offset : int, default: 0
        Number of elements to offset from the start of the buffer.
    length : int, optional
        If the buffer is a bit-mask, specifies a number of bits to read
        from the buffer. Has no effect otherwise.

    Returns
    -------
    np.ndarray

    Notes
    -----
    The returned array doesn't own the memory. The caller of this function is
    responsible for keeping the memory owner object alive as long as
    the returned NumPy array is being used.
    """
    ...

def bitmask_to_bool_ndarray(bitmask: np.ndarray, mask_length: int, first_byte_offset: int = ...) -> np.ndarray:
    """
    Convert bit-mask to a boolean NumPy array.

    Parameters
    ----------
    bitmask : np.ndarray[uint8]
        NumPy array of uint8 dtype representing the bitmask.
    mask_length : int
        Number of elements in the mask to interpret.
    first_byte_offset : int, default: 0
        Number of elements to offset from the start of the first byte.

    Returns
    -------
    np.ndarray[bool]
    """
    ...

def set_nulls(data: np.ndarray | pd.Series, col: Column, validity: tuple[Buffer, tuple[DtypeKind, int, str, str]] | None, allow_modify_inplace: bool = ...): # -> ndarray[Unknown, Unknown] | Series | NDArray[Any]:
    """
    Set null values for the data according to the column null kind.

    Parameters
    ----------
    data : np.ndarray or pd.Series
        Data to set nulls in.
    col : Column
        Column object that describes the `data`.
    validity : tuple(Buffer, dtype) or None
        The return value of ``col.buffers()``. We do not access the ``col.buffers()``
        here to not take the ownership of the memory of buffer objects.
    allow_modify_inplace : bool, default: True
        Whether to modify the `data` inplace when zero-copy is possible (True) or always
        modify a copy of the `data` (False).

    Returns
    -------
    np.ndarray or pd.Series
        Data with the nulls being set.
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from pandas._typing import NpDtype, Scalar, npt
from pandas.core.dtypes.dtypes import PandasDtype
from pandas.core.arraylike import OpsMixin
from pandas.core.arrays._mixins import NDArrayBackedExtensionArray
from pandas.core.strings.object_array import ObjectStringArrayMixin

class PandasArray(OpsMixin, NDArrayBackedExtensionArray, ObjectStringArrayMixin):
    """
    A pandas ExtensionArray for NumPy data.

    This is mostly for internal compatibility, and is not especially
    useful on its own.

    Parameters
    ----------
    values : ndarray
        The NumPy ndarray to wrap. Must be 1-dimensional.
    copy : bool, default False
        Whether to copy `values`.

    Attributes
    ----------
    None

    Methods
    -------
    None
    """
    _typ = ...
    __array_priority__ = ...
    _ndarray: np.ndarray
    _dtype: PandasDtype
    _internal_fill_value = ...
    def __init__(self, values: np.ndarray | PandasArray, copy: bool = ...) -> None:
        ...
    
    @property
    def dtype(self) -> PandasDtype:
        ...
    
    def __array__(self, dtype: NpDtype | None = ...) -> np.ndarray:
        ...
    
    def __array_ufunc__(self, ufunc: np.ufunc, method: str, *inputs, **kwargs): # -> _NotImplementedType | tuple[Unknown, ...] | Any | tuple[Self@PandasArray, ...] | Self@PandasArray | None:
        ...
    
    def isna(self) -> np.ndarray:
        ...
    
    def any(self, *, axis: int | None = ..., out=..., keepdims: bool = ..., skipna: bool = ...):
        ...
    
    def all(self, *, axis: int | None = ..., out=..., keepdims: bool = ..., skipna: bool = ...):
        ...
    
    def min(self, *, axis: int | None = ..., skipna: bool = ..., **kwargs) -> Scalar:
        ...
    
    def max(self, *, axis: int | None = ..., skipna: bool = ..., **kwargs) -> Scalar:
        ...
    
    def sum(self, *, axis: int | None = ..., skipna: bool = ..., min_count=..., **kwargs) -> Scalar:
        ...
    
    def prod(self, *, axis: int | None = ..., skipna: bool = ..., min_count=..., **kwargs) -> Scalar:
        ...
    
    def mean(self, *, axis: int | None = ..., dtype: NpDtype | None = ..., out=..., keepdims: bool = ..., skipna: bool = ...):
        ...
    
    def median(self, *, axis: int | None = ..., out=..., overwrite_input: bool = ..., keepdims: bool = ..., skipna: bool = ...):
        ...
    
    def std(self, *, axis: int | None = ..., dtype: NpDtype | None = ..., out=..., ddof=..., keepdims: bool = ..., skipna: bool = ...):
        ...
    
    def var(self, *, axis: int | None = ..., dtype: NpDtype | None = ..., out=..., ddof=..., keepdims: bool = ..., skipna: bool = ...):
        ...
    
    def sem(self, *, axis: int | None = ..., dtype: NpDtype | None = ..., out=..., ddof=..., keepdims: bool = ..., skipna: bool = ...):
        ...
    
    def kurt(self, *, axis: int | None = ..., dtype: NpDtype | None = ..., out=..., keepdims: bool = ..., skipna: bool = ...):
        ...
    
    def skew(self, *, axis: int | None = ..., dtype: NpDtype | None = ..., out=..., keepdims: bool = ..., skipna: bool = ...):
        ...
    
    def to_numpy(self, dtype: npt.DTypeLike | None = ..., copy: bool = ..., na_value: object = ...) -> np.ndarray:
        ...
    
    def __invert__(self) -> PandasArray:
        ...
    
    def __neg__(self) -> PandasArray:
        ...
    
    def __pos__(self) -> PandasArray:
        ...
    
    def __abs__(self) -> PandasArray:
        ...
    
    _arith_method = ...
    _str_na_value = ...





"""
This type stub file was generated by pyright.
"""

import pyarrow
from pandas._typing import IntervalClosedType

class ArrowPeriodType(pyarrow.ExtensionType):
    def __init__(self, freq) -> None:
        ...
    
    @property
    def freq(self): # -> Unknown:
        ...
    
    def __arrow_ext_serialize__(self) -> bytes:
        ...
    
    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized) -> ArrowPeriodType:
        ...
    
    def __eq__(self, other) -> bool:
        ...
    
    def __hash__(self) -> int:
        ...
    
    def to_pandas_dtype(self): # -> PeriodDtype | PandasExtensionDtype:
        ...
    


_period_type = ...
class ArrowIntervalType(pyarrow.ExtensionType):
    def __init__(self, subtype, closed: IntervalClosedType) -> None:
        ...
    
    @property
    def subtype(self):
        ...
    
    @property
    def closed(self) -> IntervalClosedType:
        ...
    
    def __arrow_ext_serialize__(self) -> bytes:
        ...
    
    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized) -> ArrowIntervalType:
        ...
    
    def __eq__(self, other) -> bool:
        ...
    
    def __hash__(self) -> int:
        ...
    
    def to_pandas_dtype(self): # -> IntervalDtype | PandasExtensionDtype:
        ...
    


_interval_type = ...



"""
This type stub file was generated by pyright.
"""

from contextlib import contextmanager
from typing import Callable, Iterator

"""
Helpers for configuring locale settings.

Name `localization` is chosen to avoid overlap with builtin `locale` module.
"""
@contextmanager
def set_locale(new_locale: str | tuple[str, str], lc_var: int = ...) -> Iterator[str | tuple[str, str]]:
    """
    Context manager for temporarily setting a locale.

    Parameters
    ----------
    new_locale : str or tuple
        A string of the form <language_country>.<encoding>. For example to set
        the current locale to US English with a UTF8 encoding, you would pass
        "en_US.UTF-8".
    lc_var : int, default `locale.LC_ALL`
        The category of the locale being set.

    Notes
    -----
    This is useful when you want to run a particular block of code under a
    particular locale, without globally setting the locale. This probably isn't
    thread-safe.
    """
    ...

def can_set_locale(lc: str, lc_var: int = ...) -> bool:
    """
    Check to see if we can set a locale, and subsequently get the locale,
    without raising an Exception.

    Parameters
    ----------
    lc : str
        The locale to attempt to set.
    lc_var : int, default `locale.LC_ALL`
        The category of the locale being set.

    Returns
    -------
    bool
        Whether the passed locale can be set
    """
    ...

def get_locales(prefix: str | None = ..., normalize: bool = ..., locale_getter: Callable[[], bytes] = ...) -> list[str] | None:
    """
    Get all the locales that are available on the system.

    Parameters
    ----------
    prefix : str
        If not ``None`` then return only those locales with the prefix
        provided. For example to get all English language locales (those that
        start with ``"en"``), pass ``prefix="en"``.
    normalize : bool
        Call ``locale.normalize`` on the resulting list of available locales.
        If ``True``, only locales that can be set without throwing an
        ``Exception`` are returned.
    locale_getter : callable
        The function to use to retrieve the current locales. This should return
        a string with each locale separated by a newline character.

    Returns
    -------
    locales : list of strings
        A list of locale strings that can be set with ``locale.setlocale()``.
        For example::

            locale.setlocale(locale.LC_ALL, locale_string)

    On error will return None (no locale available, e.g. Windows)

    """
    ...

